#!/usr/bin/env python
# crate_anon/nlp_manager/parse_medex.py

r"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

- MedEx-UIMA
  ... can't find Python version of MedEx (which preceded MedEx-UIMA)
  ... paper on Python version is https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2995636/
        ... uses Python NLTK
  ... see notes in Documents/CRATE directory
  ... MedEx-UIMA is in Java, and resolutely uses a file-based processing
      system; Main.java calls MedTagger.java (MedTagger.run_batch_medtag), and
      even in its core MedTagger.medtagging() function it's making files in
      directories; that's deep in the core of its NLP thinking so we can't
      change that behaviour without creating a fork.
      So the obvious way to turn this into a proper "live" pipeline would be
      for the calling code to
            fire up a receiving process - Python launching custom Java
            create its own temporary directory - Python
            receive data - Python
            stash it on disk - Python
            call the MedEx function - Python -> stdout -> custom Java -> MedEx
            return the results - custom Java signals "done" -> Python reads stdin?
            and clean up - Python
      Not terribly elegant, but might be fast enough (and almost certainly
      much faster than reloading Java regularly!).
  ... output comes from its MedTagger.print_result() function
  ... would need a per-process-unique temporary directory, since it scans all
      files in the input directory (and similarly one output directory); would
      do that in Python


MedEx-UIMA is firmly (and internally) wedded to a file-based processing
system. So we need to:

    - create a process-specific pair of temporary directories;
    - fire up a receiving process
    - pass data (1) to file and (2) signal that there's data available;
    - await a "data ready" reply and read the data from disk;
    - clean up (delete files) in readiness for next data chunk.

NOTE ALSO that MedEx's MedTagger class writes to stdout (though not
stderr). Option 1: move our logs to stdout and use stderr for signalling.
Option 2: keep things as they are and just use a stdout signal that's
not used by MedEx. Went with option 2; simpler and more consistent esp.
for logging.

How do we clean up the temporary directories?
- __del__ is not the opposite of __init__
  http://www.algorithm.co.il/blogs/programming/python-gotchas-1-__del__-is-not-the-opposite-of-__init__/
- http://eli.thegreenplace.net/2009/06/12/safely-using-destructors-in-python  # noqa

PROBLEMS:
-   NLP works fine, but UK-style abbreviations e.g. "qds" not recognized where
    "q.i.d." is. US abbreviations: e.g.
    http://www.d.umn.edu/medweb/Modules/Prescription/Abbreviations.html

    Places to look, and things to try adding:

        resources/TIMEX/norm_patterns/NormFREQword

            qds=>R1P6H

        resources/TIMEX/rules/frequency_rules

            //QID ( 4 times a day
            expression="[Qq]\.?[Ii]\.?[Dd]\.?[ ]*\((.*?)\)",val="R1P6H"

            // RNC: qds
            expression="[Qq]\.?[Dd]\.?[Ss]\.?[ ]*\((.*?)\)",val="R1P6H"

        ... looked like it was correct, but not working
        ... are this files compiled in, rather than being read live?
        ... do I have the user or the developer version?

    ... not there yet.
    Probably need to recompile. See MedEx's Readme.txt

    reference to expression/val (as in frequency_rules)
    TIMEX.Rule._add_rule()
        ... from TIMEX.Rule.Rule via a directory walker
        ... from TIMEX.ProcessingEngine.ProcessingEngine()
            ... via semi-hardcoded file location relative to class's location
                ... via rule_dir, set to .../TIMEX/rules

    Detect a file being accessed:

        sudo apt install inotify-tools
        inotifywait -m FILE

    ... frequency_rules IS opened.

    OVERALL SEQUENCE:

    org.apache.medex.Main [OR: CrateNedexPipeline.java]
    org.apache.medex.MedTagger.run_batch_medtag
    ... creeates an org.apache.NLPTools.Document
        ... not obviously doing frequency stuff, or drug recognition
    ... then runs org.apache.medex.MedTagger.medtagging(doc)
        ... this does most of the heavy lifting, I think
        ... uses ProcessingEngine freq_norm_engine
            ... org.apache.TIMEX.ProcessingEngine
            ... but it may be that this just does frequency NORMALIZATION, not frequency finding
        ... uses SemanticRuleEngine rule_engine
            ... which is org.apache.medex.SemanticRuleEngine
            ... see all the regexlist.put(..., "FREQ") calls
            ... note double-escaping \\ for Java's benefit

-   Rebuilding MedEx:

    export MEDEX_HOME=~/dev/MedEx_UIMA_1.3.6  # or similar
    cd ${MEDEX_HOME}
    # OPTIONAL # find . -name "*.class" -exec rm {} \;  # remove old compiled files
    javac \
        -classpath "${MEDEX_HOME}/src:${MEDEX_HOME}/lib/*" \
        src/org/apache/medex/Main.java \
        -d bin

    # ... will also compile dependencies

    See build_medex_itself.py

-   YES. If you add to org.apache.medex.SemanticRuleEngine, with extra entries
    in the "regexlist.put(...)" sequence, new frequencies appear in the output.

    To get them normalized as well, add them to frequency_rules.

    Specifics:
    (a) SemanticRuleEngine.java

        // EXTRA FOR UK FREQUENCIES (see http://www.evidence.nhs.uk/formulary/bnf/current/general-reference/latin-abbreviations)
        // NB case-insensitive regexes in SemanticRuleEngine.java, so ignore case here
        regexlist.put("^(q\\.?q\\.?h\\.?)( |$)", "FREQ");  // qqh, quarta quaque hora (RNC)
        regexlist.put("^(q\\.?d\\.?s\\.?)( |$)", "FREQ");  // qds, quater die sumendum (RNC); must go before existing competing expression: regexlist.put("^q(\\.|)\\d+( |$)","FREQ");
        regexlist.put("^(t\\.?d\\.?s\\.?)( |$)", "FREQ");  // tds, ter die sumendum (RNC)
        regexlist.put("^(b\\.?d\\.?)( |$)", "FREQ");  // bd, bis die (RNC)
        regexlist.put("^(o\\.?d\\.?)( |$)", "FREQ");  // od, omni die (RNC)
        regexlist.put("^(mane)( |$)", "FREQ");  // mane (RNC)
        regexlist.put("^(o\\.?m\\.?)( |$)", "FREQ");  // om, omni mane (RNC)
        regexlist.put("^(nocte)( |$)", "FREQ");  // nocte (RNC)
        regexlist.put("^(o\\.?n\\.?)( |$)", "FREQ");  // on, omni nocte (RNC)
        regexlist.put("^(fortnightly)( |$)", "FREQ");  // fortnightly (RNC)
        regexlist.put("^((?:2|two)\s+weekly)\b", "FREQ");  // fortnightly (RNC)
        regexlist.put("argh", "FREQ");  // fortnightly (RNC)
        // ALREADY IMPLEMENTED BY MedEx: tid (ter in die)
        // NECESSITY, NOT FREQUENCY: prn (pro re nata)
        // TIMING, NOT FREQUENCY: ac (ante cibum); pc (post cibum)

    (b) frequency_rules

// EXTRA FOR UK FREQUENCIES (see http://www.evidence.nhs.uk/formulary/bnf/current/general-reference/latin-abbreviations)
// NB case-sensitive regexes in Rule.java, so offer upper- and lower-case alternatives here
// qqh, quarta quaque hora (RNC)
expression="\b[Qq]\.?[Qq]\.?[Hh]\.?\b",val="R1P4H"
// qds, quater die sumendum (RNC); MUST BE BEFORE COMPETING "qd" (= per day) expression: expression="[Qq]\.?[ ]?[Dd]\.?",val="R1P24H"
expression="\b[Qq]\.?[Dd]\.?[Ss]\.?\b",val="R1P6H"
// tds, ter die sumendum (RNC)
expression="\b[Tt]\.?[Dd]\.?[Ss]\.?\b",val="R1P8H"
// bd, bis die (RNC)
expression="\b[Bb]\.?[Dd]\.?\b",val="R1P12H"
// od, omni die (RNC)
expression="\b[Oo]\.?[Dd]\.?\b",val="R1P24H"
// mane (RNC)
expression="\b[Mm][Aa][Nn][Ee]\b",val="R1P24H"
// om, omni mane (RNC)
expression="\b[Oo]\.?[Mm]\.?\b",val="R1P24H"
// nocte (RNC)
expression="\b[Nn][Oo][Cc][Tt][Ee]\b",val="R1P24H"
// on, omni nocte (RNC)
expression="\b[Oo]\.?[Nn]\.?\b",val="R1P24H"
// fortnightly and variants (RNC); unsure if TIMEX3 format is right
expression="\b[Ff][Oo][Rr][Tt][Nn][Ii][Gg][Hh][Tt][Ll][Yy]\b",val="R1P2WEEK"
expression="\b(?:2|[Tt][Ww][Oo])\s+[Ww][Ee][Ee][Kk][Ll][Yy]\b",val="R1P2WEEK"
// monthly (RNC)
expression="\b[Mm][Oo][Nn][Tt][Hh][Ll][Yy]\b",val="R1P1MONTH"
//
// ALREADY IMPLEMENTED BY MedEx: tid (ter in die)
// NECESSITY, NOT FREQUENCY: prn (pro re nata)
// TIMING, NOT FREQUENCY: ac (ante cibum); pc (post cibum)

    (c) source:

        http://www.evidence.nhs.uk/formulary/bnf/current/general-reference/latin-abbreviations

-   How about routes of administration?

        MedTagger.printResult()
            route is in FStr_list[5]
        ... called from MedTagger.medtagging()
            route is in FStr_list_final[5]
            before that, is in FStr (separated by \n)
                ... from formatDruglist
                ...
                ... from logs, appears first next to "input for tagger" at
                    which point it's in
                        sent_token_array[j] (e.g. "po")
                        sent_tag_array[j] (e.g. "RUT" = route)
                ... from tag_dict
                ... from filter_tags
                ... from (Document) doc.filtered_drug_tag()
                ...
                ... ?from MedTagger.medtagging() calling doc.add_drug_tag()
                ... no, not really; is in this bit:
                    SuffixArray sa = new SuffixArray(...);
                    Vector<SuffixArrayResult> result = sa.search();
                ... and then each element of result has a "semantic_type"
                    member that can be "RUT"
        ... SuffixArray.search()
            semantic_type=this.lex.sem_list().get(i);

            ... where lex comes from MedTagger:
                this.lex = new Lexicon(this.lex_fname);
        ... Lexicon.sem_list() returns Lexicon.semantic_list
            ... Lexicon.Lexicon() constructs using MedTagger's this.lex_fname
            ... which is lexicon.cfg

        ... aha! There it is. If a line in lexicon.cfg has a RUT tag, it'll
            appear as a route. So:
                grep "RUT$" lexicon.cfg | sort

            bedside	RUT
            by mouth	RUT
            drip	RUT
            gt	RUT
            g tube	RUT
            g-tube	RUT
            gtube	RUT
            im injection	RUT
            im	RUT
            inhalation	RUT
            inhalatn	RUT
            inhaled	RUT
            intramuscular	RUT
            intravenously	RUT
            intravenous	RUT
            iv	RUT
            j tube	RUT
            j-tube	RUT
            jtube	RUT
            nare	RUT
            nares	RUT
            naris	RUT
            neb	RUT
            nostril	RUT
            orally	RUT
            oral	RUT
            ou	RUT
            patch	DDF-DOSEUNIT-RUT
            per gt	RUT
            per mouth	RUT
            per os	RUT
            per rectum	RUT
            per tube	RUT
            p. g	RUT
            pgt	RUT
            png	RUT
            pnj	RUT
            p.o	RUT
            po	RUT
            sc	RUT
            sl	RUT
            sq	RUT
            subc	RUT
            subcu	RUT
            subcutaneously	RUT
            subcutaneous	RUT
            subcut	RUT
            subling	RUT
            sublingual	RUT
            sub q	RUT
            subq	RUT
            swallow	RUT
            swish and spit	RUT
            sw&spit	RUT
            sw&swall	RUT
            topically	RUT
            topical	RUT
            topical tp	RUT
            trans	RUT
            with spacer	RUT

    Looks like these are not using synonyms. Note also format is route\tRUT
    Note also that the first element is always forced to lower case (in
        Lexicon.Lexicon()), so presumably it's case-insensitive.
    There's no specific comment format (though any line that doesn't resolve
        to two items when split on a tab looks like it's ignored).
    So we might want to add more; use

        build_medex_itself.py --extraroutes >> lexicon.cfg

-   Note that all frequencies and routes must be in the lexicon.
    And all frequencies must be in SemanticRuleEngine.java (and, to be
    normalized, frequency_rules).

-   USEFUL BIT FOR CHECKING RESULTS:

    SELECT
        sentence_text,
        drug, generic_name,
        form, strength, dose_amount,
        route, frequency, frequency_timex3,
        duration, necessity
    FROM anonymous_output.drugs;

-   ENCODING
    - Pipe encoding (to Java's stdin, from Java's stdout) encoding is the less
      important as we're only likely to send/receive ASCII. It's hard-coded
      to UTF-8.
    - File encoding is vital and is hard-coded to UTF-8 here and in the
      receiving Java.

    - We have no direct influence over the MedTagger code for output (unless we
      modify it). The output function is MedTagger.print_result(), which
      (line 2040 of MedTagger.java) calls out.write(stuff).
      The out variable is set by
      			this.out = new BufferedWriter(new FileWriter(output_dir
					+ File.separator + doc.fname()));
      That form of the FileWriter constructor, FileWriter(String fileName),
      uses the "default character encoding", as per
        https://docs.oracle.com/javase/7/docs/api/java/io/FileWriter.html
      That default is given by System.getProperty("file.encoding").
      However, we don't have to do something daft like asking the Java to
      report its file encoding to Python through a pipe; instead, we can set
      the Java default encoding. It can't be done dynamically, but it can be
      done at JVM launch:
        http://stackoverflow.com/questions/361975/setting-the-default-java-character-encoding
      Therefore, we should have a Java parameter specified in the config file
      as
            -Dfile.encoding=UTF-8

"""  # noqa

import logging
import os
import shlex
import subprocess
import sys
import tempfile
from typing import Any, Dict, Generator, List, Optional, Tuple

from cardinal_pythonlib.rnc_lang import AttrDict
from sqlalchemy import Column, Index, Integer, String, Text

from crate_anon.common.fileops import mkdir_p
from crate_anon.nlp_manager.base_nlp_parser import BaseNlpParser
from crate_anon.nlp_manager.constants import (
    MEDEX_DATA_READY_SIGNAL,
    MEDEX_RESULTS_READY_SIGNAL,
)
from crate_anon.nlp_manager.nlp_definition import NlpDefinition

log = logging.getLogger(__name__)


DATA_FILENAME = "crate_medex.txt"
DATA_FILENAME_KEEP = "crate_medex_{}.txt"

USE_TEMP_DIRS = True
# ... True for production; False to see e.g. logs afterwards, by keeping
# everything in a subdirectory of the user's home directory (see hard-coded
# nastiness -- for debugging only)

SKIP_IF_NO_GENERIC = True
# ... Probably should be True. MedEx returns hits for drug "Thu" with no
# generic drug; this from its weekday lexicon, I think.


class Medex(BaseNlpParser):
    """Class controlling a Medex-UIMA external process, via our custom
    Java interface, CrateMedexPipeline.java.
    """

    NAME = "MedEx"

    def __init__(self,
                 nlpdef: NlpDefinition,
                 cfgsection: str,
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef, cfgsection=cfgsection, commit=commit)

        self._tablename = nlpdef.opt_str(
            cfgsection, 'desttable', required=True)

        self._max_external_prog_uses = nlpdef.opt_int(
            cfgsection, 'max_external_prog_uses', default=0)

        self._progenvsection = nlpdef.opt_str(cfgsection, 'progenvsection')
        if self._progenvsection:
            self._env = nlpdef.get_env_dict(self._progenvsection, os.environ)
        else:
            self._env = os.environ.copy()
        self._env["NLPLOGTAG"] = nlpdef.get_logtag() or '.'
        # ... because passing a "-lt" switch with no parameter will make
        # CrateGatePipeline.java complain and stop

        if USE_TEMP_DIRS:
            self._inputdir = tempfile.TemporaryDirectory()
            self._outputdir = tempfile.TemporaryDirectory()
            self._workingdir = tempfile.TemporaryDirectory()
            # ... these are autodeleted when the object goes out of scope; see
            #     https://docs.python.org/3/library/tempfile.html
            # ... which manages it using weakref.finalize
        else:
            homedir = os.path.expanduser("~")
            self._inputdir = AttrDict(
                name=os.path.join(homedir, "medextemp", "input"))
            mkdir_p(self._inputdir.name)
            self._outputdir = AttrDict(
                name=os.path.join(homedir, "medextemp", "output"))
            mkdir_p(self._outputdir.name)
            self._workingdir = AttrDict(
                name=os.path.join(homedir, "medextemp", "working"))
            mkdir_p(self._workingdir.name)

        progargs = nlpdef.opt_str(cfgsection, 'progargs', required=True)
        formatted_progargs = progargs.format(**self._env)
        self._progargs = shlex.split(formatted_progargs)
        self._progargs.extend([
            "-data_ready_signal", MEDEX_DATA_READY_SIGNAL,
            "-results_ready_signal", MEDEX_RESULTS_READY_SIGNAL,
            "-i", self._inputdir.name,
            "-o", self._outputdir.name,
        ])

        self._n_uses = 0
        self._pipe_encoding = 'utf8'
        self._file_encoding = 'utf8'
        self._p = None  # the subprocess
        self._started = False

    @classmethod
    def print_info(cls, file=sys.stdout):
        print("NLP class to talk to MedEx-UIMA, a medication-finding tool "
              "(https://www.ncbi.nlm.nih.gov/pubmed/25954575).", file=file)

    # -------------------------------------------------------------------------
    # External process control
    # -------------------------------------------------------------------------

    def _start(self) -> None:
        """
        Launch the external process.
        """
        if self._started:
            return
        args = self._progargs

        # Nasty MedEx hacks
        cwd = os.getcwd()
        log.info("for MedEx's benefit, changing to directory: {}".format(
            self._workingdir.name))
        os.chdir(self._workingdir.name)
        sentsdir = os.path.join(self._workingdir.name, "sents")
        log.info("making temporary sentences directory: {}".format(sentsdir))
        mkdir_p(sentsdir)
        logdir = os.path.join(self._workingdir.name, "log")
        log.info("making temporary log directory: {}".format(logdir))
        mkdir_p(logdir)

        log.info("launching command: {}".format(args))
        self._p = subprocess.Popen(args,
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE,
                                   # stderr=subprocess.PIPE,
                                   shell=False,
                                   bufsize=1)
        # ... don't ask for stderr to be piped if you don't want it; firstly,
        # there's a risk that if you don't consume it, something hangs, and
        # secondly if you don't consume it, you see it on the console, which is
        # helpful.
        self._started = True
        log.info("returning to working directory {}".format(cwd))
        os.chdir(cwd)

    def _encode_to_subproc_stdin(self, text: str) -> None:
        """Send text to the external program (via its stdin), encoding it in
        the process (typically to UTF-8)."""
        log.debug("SENDING: " + text)
        bytes_ = text.encode(self._pipe_encoding)
        self._p.stdin.write(bytes_)

    def _flush_subproc_stdin(self) -> None:
        """Flushes what we're sending to the external program via its stdin."""
        self._p.stdin.flush()

    def _decode_from_subproc_stdout(self) -> str:
        """Translate what we've received from the external program's stdout,
        from its specific encoding (usually UTF-8) to a Python string."""
        bytes_ = self._p.stdout.readline()
        text = bytes_.decode(self._pipe_encoding)
        log.debug("RECEIVING: " + repr(text))
        return text

    def _finish(self) -> None:
        """
        Close down the external process.
        """
        if not self._started:
            return
        self._p.communicate()  # close p.stdout, wait for the subprocess to exit
        self._started = False

    def _signal_data_ready(self) -> bool:
        """Returns: OK?"""
        if self._finished():
            return False
        self._encode_to_subproc_stdin(MEDEX_DATA_READY_SIGNAL + os.linesep)
        self._flush_subproc_stdin()
        return True

    def _await_results_ready(self) -> bool:
        """Returns: ok?"""
        while True:
            if self._finished():
                return False
            line = self._decode_from_subproc_stdout()
            if line == MEDEX_RESULTS_READY_SIGNAL + os.linesep:
                return True

    def _finished(self) -> bool:
        if not self._started:
            return True
        self._p.poll()
        finished = self._p.returncode is not None
        if finished:
            self._started = False
        return finished

    # -------------------------------------------------------------------------
    # Input processing
    # -------------------------------------------------------------------------

    def parse(self, text: str) -> Generator[Tuple[str, Dict[str, Any]],
                                            None, None]:
        """
        - Send text to the external process, and receive the result.
        - Note that associated data is not passed into this function, and is
          kept in the Python environment, so we can't run into any problems
          with the transfer to/from the Java program garbling important data.
          All we send to the subprocess is the text (and an input_terminator).
          Then, we may receive MULTIPLE sets of data back ("your text contains
          the following 7 people/drug references/whatever"), followed
          eventually by the output_terminator, at which point this set is
          complete.
        """
        self._n_uses += 1
        self._start()  # ensure started
        if USE_TEMP_DIRS:
            basefilename = DATA_FILENAME
        else:
            basefilename = DATA_FILENAME_KEEP.format(self._n_uses)
        inputfilename = os.path.join(self._inputdir.name, basefilename)
        outputfilename = os.path.join(self._outputdir.name, basefilename)
        # ... MedEx gives output files the SAME NAME as input files.

        with open(inputfilename, mode='w',
                  encoding=self._file_encoding) as infile:
            # log.critical("text: {}".format(repr(text)))
            infile.write(text)

        if (not self._signal_data_ready() or  # send
                not self._await_results_ready()):  # receive
            log.critical("Subprocess terminated unexpectedly")
            os.remove(inputfilename)
            # We were using "log.critical()" and "return", but if the Medex
            # processor is misconfigured, the failed processor can be run over
            # thousands of records over many hours before the failure is
            # obvious. Changed 2017-03-17.
            raise ValueError("Java interface to Medex failed - miconfigured?")

        with open(outputfilename, mode='r',
                  encoding=self._file_encoding) as infile:
            resultlines = infile.readlines()
        for line in resultlines:
            # log.critical("received: {}".format(line))
            # Output code, from MedTagger.print_result():
            # out.write(
            #     index + 1 + "\t" + sent_text + "|" + drug + "|" + brand + "|"
            #     + dose_form + "|" + strength + "|" + dose_amt + "|" + route
            #     + "|" + frequency + "|" + duration + "|" + necessity + "|"
            #     + umls_code + "|" + rx_code + "|" + generic_code + "|" + generic_name + "\n");  # noqa
            # NOTE that the text can contain | characters. So work from the
            # right.
            line = line.rstrip()  # remove any trailing newline
            fields = line.split('|')
            if len(fields) < 14:
                log.warning("Bad result received: {}".format(repr(line)))
                continue
            generic_name = self.str_or_none(fields[-1])
            if not generic_name and SKIP_IF_NO_GENERIC:
                continue
            generic_code = self.int_or_none(fields[-2])
            rx_code = self.int_or_none(fields[-3])
            umls_code = self.str_or_none(fields[-4])
            (necessity, necessity_startpos, necessity_endpos) = \
                self.get_text_start_end(fields[-5])
            (duration, duration_startpos, duration_endpos) = \
                self.get_text_start_end(fields[-6])
            (_freq_text, frequency_startpos, frequency_endpos) = \
                self.get_text_start_end(fields[-7])
            frequency, frequency_timex = self.frequency_and_timex(_freq_text)
            (route, route_startpos, route_endpos) = \
                self.get_text_start_end(fields[-8])
            (dose_amount, dose_amount_startpos, dose_amount_endpos) = \
                self.get_text_start_end(fields[-9])
            (strength, strength_startpos, strength_endpos) = \
                self.get_text_start_end(fields[-10])
            (form, form_startpos, form_endpos) = \
                self.get_text_start_end(fields[-11])
            (brand, brand_startpos, brand_endpos) = \
                self.get_text_start_end(fields[-12])
            (drug, drug_startpos, drug_endpos) = \
                self.get_text_start_end(fields[-13])
            _start_bit = '|'.join(fields[0:-13])
            _index_text, sent_text = _start_bit.split('\t', maxsplit=1)
            index = self.int_or_none(_index_text)
            yield self._tablename, {
                'sentence_index': index,
                'sentence_text': sent_text,

                'drug': drug,
                'drug_startpos': drug_startpos,
                'drug_endpos': drug_endpos,

                'brand': brand,
                'brand_startpos': brand_startpos,
                'brand_endpos': brand_endpos,

                'form': form,
                'form_startpos': form_startpos,
                'form_endpos': form_endpos,

                'strength': strength,
                'strength_startpos': strength_startpos,
                'strength_endpos': strength_endpos,

                'dose_amount': dose_amount,
                'dose_amount_startpos': dose_amount_startpos,
                'dose_amount_endpos': dose_amount_endpos,

                'route': route,
                'route_startpos': route_startpos,
                'route_endpos': route_endpos,

                'frequency': frequency,
                'frequency_startpos': frequency_startpos,
                'frequency_endpos': frequency_endpos,
                'frequency_timex3': frequency_timex,

                'duration': duration,
                'duration_startpos': duration_startpos,
                'duration_endpos': duration_endpos,

                'necessity': necessity,
                'necessity_startpos': necessity_startpos,
                'necessity_endpos': necessity_endpos,

                'umls_code': umls_code,
                'rx_code': rx_code,
                'generic_code': generic_code,
                'generic_name': generic_name,
            }

        # Since MedEx scans all files in the input directory, then if we're
        # not using temporary directories (and are therefore using a new
        # filename per item), we should remove the old one.
        os.remove(inputfilename)

        # Restart subprocess?
        if (self._max_external_prog_uses > 0 and
                self._n_uses % self._max_external_prog_uses == 0):
            log.info("relaunching app after {} uses".format(
                self._max_external_prog_uses))
            self._finish()
            self._start()

    @staticmethod
    def get_text_start_end(medex_str: Optional[str]) -> Tuple[Optional[str],
                                                              Optional[int],
                                                              Optional[int]]:
        """
        MedEx returns 'drug', 'strength', etc. as "aspirin[7,14]", where the
        text is followed by the start position (zero-indexed) and the end
        position (one beyond the last character) (zero-indexed).
        """
        if not medex_str:
            return None, None, None
        lbracket = medex_str.rfind('[')  # -1 for not found
        comma = medex_str.rfind(',')
        rbracket = medex_str.rfind(']')
        try:
            if lbracket == -1 or not (lbracket < comma < rbracket):
                raise ValueError()
            text = medex_str[:lbracket]
            lpos = int(medex_str[lbracket + 1:comma])
            rpos = int(medex_str[comma + 1:rbracket])
            return text, lpos, rpos
        except (TypeError, ValueError):
            log.warning("Bad string[left, right] format: {}".format(
                repr(medex_str)))
            return None, None, None

    @staticmethod
    def int_or_none(text: Optional[str]) -> Optional[int]:
        try:
            return int(text)
        except (TypeError, ValueError):
            return None

    @staticmethod
    def str_or_none(text: Optional[str]) -> Optional[str]:
        return None if not text else text

    @staticmethod
    def frequency_and_timex(text: str) -> Tuple[Optional[str], Optional[str]]:
        """
        Splits e.g. b.i.d.(R1P12H)
        """
        if not text:
            return None, None
        lbracket = text.rfind('(')
        rbracket = text.rfind(')')
        if (lbracket == -1 or
                not (lbracket < rbracket) or
                rbracket != len(text) - 1):
            return None, None
        return text[0:lbracket], text[lbracket + 1:rbracket]

    # -------------------------------------------------------------------------
    # Test
    # -------------------------------------------------------------------------

    def test(self, verbose: bool=False) -> None:
        """
        Test the send function.
        """
        self.test_parser([
            "Bob Hope visited Seattle and took venlafaxine M/R 375mg od.",
            "James Joyce wrote Ulysses whilst taking aspirin 75mg mane."
        ])

    # -------------------------------------------------------------------------
    # Database structure
    # -------------------------------------------------------------------------

    def dest_tables_columns(self) -> Dict[str, List[Column]]:
        # https://phekb.org/sites/phenotype/files/MedEx_UIMA_eMERGE_short.pdf
        # RxNorm: https://www.nlm.nih.gov/research/umls/rxnorm/overview.html
        # UMLS: https://www.nlm.nih.gov/research/umls/new_users/glossary.html
        # UMLS CUI max length: https://www.nlm.nih.gov/research/umls/knowledge_sources/metathesaurus/release/columns_data_elements.html  # noqa
        # TIMEX3:
        # http://www.timeml.org/tempeval2/tempeval2-trial/guidelines/timex3guidelines-072009.pdf  # noqa
        # http://www.timeml.org/publications/timeMLdocs/timeml_1.2.1.html#timex3  # noqa
        drug_length = 50  # guess
        form_length = 50  # guess; "Powder For Oral Suspension" (26) is one
        strength_length = 25  # guess
        dose_amount_length = 25  # guess
        route_length = 25  # guess
        frequency_length = 30  # guess
        timex3_length = 30  # guess
        duration_length = 30  # guess
        necessity_length = 30  # guess
        umls_cui_max_length = 8  # definite
        startposdef = "Start position (zero-based) of "
        endposdef = (
            "End position (zero-based index of one beyond last character) of ")
        return {
            self._tablename: [
                Column('sentence_index', Integer,
                       doc="One-based index of sentence in text"),
                Column('sentence_text', Text,
                       doc="Text recognized as a sentence by MedEx"),

                Column('drug', String(drug_length),
                       doc="Drug name, as in the text"),
                Column('drug_startpos', Integer,
                       doc=startposdef + "drug"),
                Column('drug_endpos', Integer,
                       doc=endposdef + "drug"),

                Column('brand', String(drug_length),
                       doc="Drug brand name (?lookup ?only if given)"),
                Column('brand_startpos', Integer,
                       doc=startposdef + "brand"),
                Column('brand_endpos', Integer,
                       doc=endposdef + "brand"),

                Column('form', String(form_length),
                       doc="Drug/dose form (e.g. 'tablet')"),
                Column('form_startpos', Integer,
                       doc=startposdef + "form"),
                Column('form_endpos', Integer,
                       doc=endposdef + "form"),

                Column('strength', String(strength_length),
                       doc="Strength (e.g. '75mg')"),
                Column('strength_startpos', Integer,
                       doc=startposdef + "strength"),
                Column('strength_endpos', Integer,
                       doc=endposdef + "strength"),

                Column('dose_amount', String(dose_amount_length),
                       doc="Dose amount (e.g. '2 tablets')"),
                Column('dose_amount_startpos', Integer,
                       doc=startposdef + "dose_amount"),
                Column('dose_amount_endpos', Integer,
                       doc=endposdef + "dose_amount"),

                Column('route', String(route_length),
                       doc="Route (e.g. 'by mouth')"),
                Column('route_startpos', Integer,
                       doc=startposdef + "route"),
                Column('route_endpos', Integer,
                       doc=endposdef + "route"),

                Column('frequency', String(frequency_length),
                       doc="Frequency (e.g. 'b.i.d.')"),
                Column('frequency_startpos', Integer,
                       doc=startposdef + "frequency"),
                Column('frequency_endpos', Integer,
                       doc=endposdef + "frequency"),
                Column('frequency_timex3', String(timex3_length),
                       doc="Normalized frequency in TIMEX3 format "
                           "(e.g. 'R1P12H')"),

                Column('duration', String(duration_length),
                       doc="Duration (e.g. 'for 10 days')"),
                Column('duration_startpos', Integer,
                       doc=startposdef + "duration"),
                Column('duration_endpos', Integer,
                       doc=endposdef + "duration"),

                Column('necessity', String(necessity_length),
                       doc="Necessity (e.g. 'prn')"),
                Column('necessity_startpos', Integer,
                       doc=startposdef + "necessity"),
                Column('necessity_endpos', Integer,
                       doc=endposdef + "necessity"),

                Column('umls_code', String(umls_cui_max_length),
                       doc="UMLS CUI"),
                Column('rx_code', Integer,
                       doc="RxNorm RxCUI for drug"),
                Column('generic_code', Integer,
                       doc="RxNorm RxCUI for generic name"),
                Column('generic_name', String(drug_length),
                       doc="Generic drug name (associated with RxCUI code)"),
            ]
        }

    def dest_tables_indexes(self) -> Dict[str, List[Index]]:
        # return {
        #     self._tablename: [
        #         Index('idx_generic_name', 'generic_name'),
        #     ]
        # }
        return {}



#!/usr/bin/env python
# crate_anon/nlp_manager/build_medex_java_interface.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Script to compile Java source for CrateMedexPipeline
"""

import argparse
import logging
import os
import subprocess
import tempfile

# from crate_anon.common.fileops import moveglob, rmglob
from crate_anon.common.logsupport import configure_logger_for_colour
from crate_anon.nlp_manager.constants import (
    MEDEX_PIPELINE_CLASSNAME,
    MEDEX_DATA_READY_SIGNAL,
    MEDEX_RESULTS_READY_SIGNAL,
)


log = logging.getLogger(__name__)

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DEFAULT_BUILD_DIR = os.path.join(THIS_DIR, 'compiled_nlp_classes')
SOURCE_FILE = os.path.join(THIS_DIR, MEDEX_PIPELINE_CLASSNAME + '.java')
DEFAULT_MEDEX_DIR = os.path.join(os.path.expanduser('~'), 'dev',
                                 'Medex_UIMA_1.3.6')
DEFAULT_JAVA = 'java'
DEFAULT_JAVAC = 'javac'


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Compile Java classes for CRATE's interface to MedEx-UIMA")
    parser.add_argument(
        '--builddir', default=DEFAULT_BUILD_DIR,
        help="Output directory for compiled .class files (default: {})".format(
            DEFAULT_BUILD_DIR))
    parser.add_argument(
        '--medexdir', default=DEFAULT_MEDEX_DIR,
        help="Root directory of MedEx installation (default: {})".format(
            DEFAULT_MEDEX_DIR))
    parser.add_argument(
        '--java', default=DEFAULT_JAVA,
        help="Java executable (default: {})".format(DEFAULT_JAVA))
    parser.add_argument(
        '--javac', default=DEFAULT_JAVAC,
        help="Java compiler (default: {})".format(DEFAULT_JAVAC))
    parser.add_argument('--verbose', '-v', action='count', default=0,
                        help="Be verbose (use twice for extra verbosity)")
    parser.add_argument(
        '--launch', action='store_true',
        help="Launch script in demonstration mode (having previously "
             "compiled it)")
    args = parser.parse_args()

    loglevel = logging.DEBUG if args.verbose >= 1 else logging.INFO
    rootlogger = logging.getLogger()
    configure_logger_for_colour(rootlogger, level=loglevel)

    medexclasses = os.path.join(args.medexdir, 'bin')
    medexlibjars = os.path.join(args.medexdir, 'lib', '*')
    classpath = os.pathsep.join([args.builddir, medexclasses, medexlibjars])
    classpath_options = ['-classpath', classpath]

    if args.launch:
        inputdir = tempfile.TemporaryDirectory()
        outputdir = tempfile.TemporaryDirectory()
        prog_args = [
            "-data_ready_signal", MEDEX_DATA_READY_SIGNAL,
            "-results_ready_signal", MEDEX_RESULTS_READY_SIGNAL,
            "-i", inputdir.name,
            "-o", outputdir.name,
        ]
        if args.verbose > 0:
            prog_args += ['-v', '-v']
        cmdargs = (
            [args.java] +
            classpath_options +
            [MEDEX_PIPELINE_CLASSNAME] +
            prog_args
        )
        log.info("Executing command: {}".format(cmdargs))
        subprocess.check_call(cmdargs)
    else:
        os.makedirs(args.builddir, exist_ok=True)
        cmdargs = (
            [args.javac, '-Xlint:unchecked'] +
            (['-verbose'] if args.verbose > 0 else []) +
            classpath_options +
            ['-d', args.builddir] +
            [SOURCE_FILE]
        )
        log.info("Executing command: {}".format(cmdargs))
        subprocess.check_call(cmdargs)
        log.info("Output *.class files are in {}".format(args.builddir))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-12 01:37
from __future__ import unicode_literals

import crate_anon.common.jsonfunc
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('userprofile', '0005_userprofile_patient_multiquery_scratchpad'),
    ]

    operations = [
        migrations.AlterField(
            model_name='userprofile',
            name='patient_multiquery_scratchpad',
            field=crate_anon.common.jsonfunc.JsonClassField(null=True, verbose_name='PatientMultiQuery scratchpad (in JSON) for builder'),  # noqa
        ),
    ]



#!/usr/bin/env python
# crate_anon/crateweb/research/forms.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import datetime
import logging
from typing import Any, Optional

from django import forms
from django.conf import settings
from django.forms import (
    BooleanField,
    CharField,
    DateField,
    FileField,
    FloatField,
    IntegerField,
    ModelForm,
)
from crate_anon.crateweb.extra.forms import (
    MultipleIntAreaField,
    MultipleWordAreaField,
)
from crate_anon.crateweb.research.models import Highlight, Query
from crate_anon.common.sql import (
    SQL_OPS_MULTIPLE_VALUES,
    SQL_OPS_VALUE_UNNECESSARY,
    QB_DATATYPE_DATE,
    QB_DATATYPE_FLOAT,
    QB_DATATYPE_INTEGER,
    QB_DATATYPE_UNKNOWN,
    QB_STRING_TYPES,
)

log = logging.getLogger(__name__)


class AddQueryForm(ModelForm):
    class Meta:
        model = Query
        fields = ['sql']
        widgets = {
            'sql': forms.Textarea(attrs={'rows': 20, 'cols': 80}),
        }


class BlankQueryForm(ModelForm):
    class Meta:
        model = Query
        fields = []


class AddHighlightForm(ModelForm):
    class Meta:
        model = Highlight
        fields = ['colour', 'text']


class BlankHighlightForm(ModelForm):
    class Meta:
        model = Highlight
        fields = []


class PidLookupForm(forms.Form):
    trids = MultipleIntAreaField(
        label='{} (TRID)'.format(settings.SECRET_MAP['TRID_FIELD']),
        required=False)
    rids = MultipleWordAreaField(
        label='{} (RID)'.format(settings.SECRET_MAP['RID_FIELD']),
        required=False)
    mrids = MultipleWordAreaField(
        label='{} (MRID)'.format(settings.SECRET_MAP['MASTER_RID_FIELD']),
        required=False)


class SQLHelperTextAnywhereForm(forms.Form):
    fkname = CharField(label="Field name containing patient research ID",
                       required=True)
    min_length = IntegerField(
        label="Minimum length of textual field (suggest e.g. 50)",
        min_value=1, required=True)
    fragment = CharField(label="String fragment to find", required=True)
    use_fulltext_index = BooleanField(
        label="Use full-text indexing where available "
        "(faster, but requires whole words)",
        required=False)
    include_content = BooleanField(
        label="Include content from fields where found (slower)",
        required=False)


def html_form_date_to_python(text: str) -> datetime.datetime:
    return datetime.datetime.strptime(text, "%Y-%m-%d")


def int_validator(text: str) -> str:
    return str(int(text))  # may raise ValueError, TypeError


def float_validator(text: str) -> str:
    return str(float(text))  # may raise ValueError, TypeError


class QueryBuilderForm(forms.Form):
    # See also querybuilder.js

    database = CharField(label="Schema", required=False)
    schema = CharField(label="Schema", required=True)
    table = CharField(label="Table", required=True)
    column = CharField(label="Column", required=True)
    datatype = CharField(label="Data type", required=True)
    offer_where = BooleanField(label="Offer WHERE?", required=False)
    # BooleanField generally needs "required=False", or you can't have False!
    where_op = CharField(label="WHERE comparison", required=False)

    date_value = DateField(label="Date value (e.g. 1900-01-31)",
                           required=False)
    int_value = IntegerField(label="Integer value", required=False)
    float_value = FloatField(label="Float value", required=False)
    string_value = CharField(label="String value", required=False)
    file = FileField(label="File (for IN)", required=False)

    def __init__(self, *args, **kwargs) -> None:
        self.file_values_list = []
        super().__init__(*args, **kwargs)

    def get_datatype(self) -> Optional[str]:
        return self.data.get('datatype', None)

    def is_datatype_unknown(self) -> bool:
        return self.get_datatype() == QB_DATATYPE_UNKNOWN

    def offering_where(self) -> bool:
        if self.is_datatype_unknown():
            return False
        return self.data.get('offer_where', False)

    def get_value_fieldname(self) -> str:
        datatype = self.get_datatype()
        if datatype == QB_DATATYPE_INTEGER:
            return "int_value"
        if datatype == QB_DATATYPE_FLOAT:
            return "float_value"
        if datatype == QB_DATATYPE_DATE:
            return "date_value"
        if datatype in QB_STRING_TYPES:
            return "string_value"
        if datatype == QB_DATATYPE_UNKNOWN:
            return ""
        raise ValueError("Invalid field type")

    def get_cleaned_where_value(self) -> Any:
        # Only call this if you've already cleaned/validated the form!
        return self.cleaned_data[self.get_value_fieldname()]

    def clean(self) -> None:
        # Check the WHERE information is sufficient.
        if 'submit_select' in self.data or 'submit_select_star' in self.data:
            # Form submitted via the "Add" method, so no checks required.
            # http://stackoverflow.com/questions/866272/how-can-i-build-multiple-submit-buttons-django-form  # noqa
            return
        if not self.offering_where():
            return
        cleaned_data = super().clean()
        if not cleaned_data['where_op']:
            self.add_error('where_op',
                           forms.ValidationError("Must specify comparison"))

        # No need for a value for NULL-related comparisons. But otherwise:
        where_op = cleaned_data['where_op']
        if where_op not in SQL_OPS_VALUE_UNNECESSARY + SQL_OPS_MULTIPLE_VALUES:
            # Can't take 0 or many parameters, so need the standard single
            # value:
            value_fieldname = self.get_value_fieldname()
            value = cleaned_data.get(value_fieldname)
            if not value:
                self.add_error(
                    value_fieldname,
                    forms.ValidationError("Must specify WHERE condition"))

        # ---------------------------------------------------------------------
        # Special processing for file upload operations
        # ---------------------------------------------------------------------
        if where_op not in SQL_OPS_MULTIPLE_VALUES:
            return
        fileobj = cleaned_data['file']
        # ... is an instance of InMemoryUploadedFile
        if not fileobj:
            self.add_error('file', forms.ValidationError("Must specify file"))
            return

        datatype = self.get_datatype()
        if datatype in QB_STRING_TYPES:
            form_to_python_fn = str
        elif datatype == QB_DATATYPE_DATE:
            form_to_python_fn = html_form_date_to_python
        elif datatype == QB_DATATYPE_INTEGER:
            form_to_python_fn = int_validator
        elif datatype == QB_DATATYPE_FLOAT:
            form_to_python_fn = float_validator
        else:
            # Safe defaults
            form_to_python_fn = str
        # Or: http://www.dabeaz.com/generators/Generators.pdf
        self.file_values_list = []
        for line in fileobj.read().decode("utf8").splitlines():
            raw_item = line.strip()
            if not raw_item or raw_item.startswith('#'):
                continue
            try:
                value = form_to_python_fn(raw_item)
            except (TypeError, ValueError):
                self.add_error('file', forms.ValidationError(
                    "File contains bad value: {}".format(repr(raw_item))))
                return
            self.file_values_list.append(value)
        if not self.file_values_list:
            self.add_error('file', forms.ValidationError(
                "No values found in file"))


class ManualPeQueryForm(forms.Form):
    sql = CharField(
        required=False,
        widget=forms.Textarea(attrs={'rows': 20, 'cols': 80})
    )



#!/usr/bin/env python
# crate_anon/crateweb/extra/forms.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from typing import List

from django import forms

from crate_anon.common.nhs import is_valid_nhs_number


# =============================================================================
# Multiple values from a text area
# =============================================================================

def clean_int(x) -> int:
    try:
        return int(x)
    except ValueError:
        raise forms.ValidationError(
            "Cannot convert to integer: {}".format(repr(x)))


def clean_nhs_number(x) -> int:
    try:
        x = int(x)
        if not is_valid_nhs_number(x):
            raise ValueError
        return x
    except ValueError:
        raise forms.ValidationError(
            "Not a valid NHS number: {}".format(repr(x)))


class MultipleIntAreaField(forms.Field):
    # See also http://stackoverflow.com/questions/29303902/django-form-with-list-of-integers  # noqa
    widget = forms.Textarea

    def clean(self, value) -> List[int]:
        return [clean_int(x) for x in value.split()]


class MultipleNhsNumberAreaField(forms.Field):
    widget = forms.Textarea

    def clean(self, value) -> List[int]:
        return [clean_nhs_number(x) for x in value.split()]


class MultipleWordAreaField(forms.Field):
    widget = forms.Textarea

    def clean(self, value) -> List[str]:
        return value.split()


class SingleNhsNumberField(forms.IntegerField):
    def clean(self, value) -> int:
        return clean_nhs_number(value)



#!/usr/bin/env python
# crate_anon/common/timing.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

This isn't very elegant, as it uses a global timing record.
"""

from collections import OrderedDict
import datetime
import logging
from cardinal_pythonlib.rnc_datetime import get_now_utc
log = logging.getLogger(__name__)


class MultiTimer(object):
    """Mutually exclusive timing of a set of events."""
    def __init__(self) -> None:
        self._timing = False
        self._overallstart = get_now_utc()
        self._starttimes = OrderedDict()  # name: start time
        self._totaldurations = OrderedDict()  # name: duration
        self._count = OrderedDict()  # name: count
        self._stack = []  # list of names

    def reset(self) -> None:
        self._overallstart = get_now_utc()
        self._starttimes.clear()
        self._totaldurations.clear()
        self._count.clear()
        self._stack.clear()

    def set_timing(self, timing: bool, reset: bool = False) -> None:
        self._timing = timing
        if reset:
            self.reset()

    def start(self, name: str, increment_count: bool = True) -> None:
        if not self._timing:
            return
        now = get_now_utc()

        # If we were already timing something else, pause that.
        if self._stack:
            last = self._stack[-1]
            self._totaldurations[last] += now - self._starttimes[last]

        # Start timing our new thing
        if name not in self._starttimes:
            self._totaldurations[name] = datetime.timedelta()
            self._count[name] = 0
        self._starttimes[name] = now
        if increment_count:
            self._count[name] += 1
        self._stack.append(name)

    def stop(self, name: str) -> None:
        if not self._timing:
            return
        now = get_now_utc()

        # Validity check
        if not self._stack:
            raise AssertionError("MultiTimer.stop() when nothing running")
        if self._stack[-1] != name:
            raise AssertionError(
                "MultiTimer.stop({}) when {} is running".format(
                    repr(name), repr(self._stack[-1])))

        # Finish what we were asked to
        self._totaldurations[name] += now - self._starttimes[name]
        self._stack.pop()

        # Now, if we were timing something else before we started "name",
        # resume...
        if self._stack:
            last = self._stack[-1]
            self._starttimes[last] = now

    def report(self) -> None:
        """Finish and report to the log."""
        while self._stack:
            self.stop(self._stack[-1])
        now = get_now_utc()
        grand_total = datetime.timedelta()
        overall_duration = now - self._overallstart
        for name, duration in self._totaldurations.items():
            grand_total += duration

        log.info("Timing summary:")
        summaries = []
        for name, duration in self._totaldurations.items():
            n = self._count[name]
            total_sec = duration.total_seconds()
            mean = total_sec / n if n > 0 else None

            summaries.append({
                'total': total_sec,
                'description': (
                    "- {}: {:.3f} s ({:.2f}%, n={}, mean={:.3f}s)".format(
                        name,
                        total_sec,
                        (100 * total_sec / grand_total.total_seconds()),
                        n,
                        mean)),
            })
        summaries.sort(key=lambda x: x['total'], reverse=True)
        for s in summaries:
            # noinspection PyTypeChecker
            log.info(s["description"])
        if not self._totaldurations:
            log.info("<no timings recorded>")

        unmetered = overall_duration - grand_total
        log.info("Unmetered time: {:.3f} s ({:.2f}%)".format(
            unmetered.total_seconds(),
            100 * unmetered.total_seconds() / overall_duration.total_seconds()
        ))


class MultiTimerContext(object):
    def __init__(self, multitimer: MultiTimer, name: str) -> None:
        self.timer = multitimer
        self.name = name

    def __enter__(self):
        self.timer.start(self.name)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.timer.stop(self.name)


timer = MultiTimer()



#!/usr/bin/env python
# crate_anon/common/sql.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import argparse
import datetime
import functools
import logging
import re
from typing import Any, Dict, Iterable, List, Tuple, Union

from pyparsing import ParseResults
from sqlalchemy import inspect
from sqlalchemy.dialects.mssql.base import MS_2012_VERSION
from sqlalchemy.engine import Engine
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.orm.session import Session
from sqlalchemy.schema import Column, Table

from crate_anon.common.formatting import sizeof_fmt
from crate_anon.common.jsonfunc import (
    METHOD_PROVIDES_INIT_KWARGS,
    METHOD_STRIP_UNDERSCORE,
    register_for_json,
)
from crate_anon.common.lang import (
    mapped_repr_stripping_underscores,
    unique_list,
)
from crate_anon.common.logsupport import main_only_quicksetup_rootlogger
from crate_anon.common.timing import MultiTimerContext, timer
from crate_anon.common.stringfunc import get_spec_match_regex
from crate_anon.common.sql_grammar import SqlGrammar, text_from_parsed
from crate_anon.common.sql_grammar_factory import (
    DIALECT_MYSQL,
    make_grammar,
    mysql_grammar,
)
from crate_anon.common.sqla import column_creation_ddl, count_star

log = logging.getLogger(__name__)

# =============================================================================
# Types
# =============================================================================

SqlArgsTupleType = Tuple[str, List[Any]]


# =============================================================================
# Constants
# =============================================================================

TIMING_COMMIT = "commit"

SQL_OPS_VALUE_UNNECESSARY = ['IS NULL', 'IS NOT NULL']
SQL_OPS_MULTIPLE_VALUES = ['IN', 'NOT IN']

SQLTYPES_INTEGER = [
    "INT", "INTEGER",
    "TINYINT", "SMALLINT", "MEDIUMINT", "BIGINT",
    "BIT", "BOOL", "BOOLEAN",
]
SQLTYPES_FLOAT = [
    "DOUBLE", "FLOAT", "DEC", "DECIMAL",
]
SQLTYPES_TEXT = [
    "CHAR", "VARCHAR", "NVARCHAR",
    "TINYTEXT", "TEXT", "NTEXT", "MEDIUMTEXT", "LONGTEXT",
]
SQLTYPES_WITH_DATE = [
    "DATE", "DATETIME", "TIMESTAMP",
]
# SQLTYPES_BINARY = [
#     "BINARY", "BLOB", "IMAGE", "LONGBLOB", "VARBINARY",
# ]

# Must match querybuilder.js:
QB_DATATYPE_INTEGER = "int"
QB_DATATYPE_FLOAT = "float"
QB_DATATYPE_DATE = "date"
QB_DATATYPE_STRING = "string"
QB_DATATYPE_STRING_FULLTEXT = "string_fulltext"
QB_DATATYPE_UNKNOWN = "unknown"
QB_STRING_TYPES = [QB_DATATYPE_STRING, QB_DATATYPE_STRING_FULLTEXT]

COLTYPE_WITH_ONE_INTEGER_REGEX = re.compile(r"^([A-z]+)\((\d+)\)$")
# ... start, group(alphabetical), literal (, group(digit), literal ), end


# =============================================================================
# SQL elements: literals
# =============================================================================

def sql_string_literal(text: str) -> str:
    # ANSI SQL: http://www.contrib.andrew.cmu.edu/~shadow/sql/sql1992.txt
    # <character string literal>
    return "'" + text.replace("'", "''") + "'"


def sql_date_literal(dt: datetime.datetime) -> str:
    # ANSI SQL: http://www.contrib.andrew.cmu.edu/~shadow/sql/sql1992.txt
    # <date string>
    return dt.strftime("'%Y-%m-%d'")


def sql_datetime_literal(dt: datetime.datetime,
                         subsecond: bool = False) -> str:
    # ANSI SQL: http://www.contrib.andrew.cmu.edu/~shadow/sql/sql1992.txt
    # <timestamp string>
    # ... the subsecond part is non-ANSI
    fmt = "'%Y-%m-%d %H:%M:%S{}'".format(".%f" if subsecond else "")
    return dt.strftime(fmt)


# def combine_db_schema_table(db: Optional[str],
#                             schema: Optional[str],
#                             table: str) -> str:
#     # ANSI SQL: http://www.contrib.andrew.cmu.edu/~shadow/sql/sql1992.txt
#     # <table name>, <qualified name>
#     if not table:
#         raise ValueError("Missing table supplied to combine_db_schema_table")
#     return ".".join(x for x in [db, schema, table] if x)


# =============================================================================
# SQL elements: identifiers
# =============================================================================

@register_for_json(method=METHOD_STRIP_UNDERSCORE)
@functools.total_ordering
class SchemaId(object):
    def __init__(self, db: str = '', schema: str = '') -> None:
        self._db = db
        self._schema = schema

    def __bool__(self) -> bool:
        return bool(self._schema)

    def __eq__(self, other: 'SchemaId') -> bool:
        return (  # ordering is for speed
            self._schema == other._schema and
            self._db == other._db
        )

    def __lt__(self, other: 'SchemaId') -> bool:
        return (
            (self._db, self._schema) <
            (other._db, other._schema)
        )

    def __hash__(self) -> int:
        return hash(str(self))

    def identifier(self, grammar: SqlGrammar) -> str:
        return make_identifier(grammar,
                               database=self._db,
                               schema=self._schema)

    def table_id(self, table: str) -> 'TableId':
        return TableId(db=self._db, schema=self._schema, table=table)

    def column_id(self, table: str, column: str) -> 'ColumnId':
        return ColumnId(db=self._db, schema=self._schema,
                        table=table, column=column)

    def db(self) -> str:
        return self._db

    def schema(self) -> str:
        return self._schema

    def __str__(self) -> str:
        return self.identifier(mysql_grammar)  # specific one unimportant

    def __repr__(self) -> str:
        return mapped_repr_stripping_underscores(
            self, ['_db', '_schema'])


@register_for_json(method=METHOD_STRIP_UNDERSCORE)
@functools.total_ordering
class TableId(object):
    def __init__(self, db: str = '', schema: str = '',
                 table: str = '') -> None:
        self._db = db
        self._schema = schema
        self._table = table

    def __bool__(self) -> bool:
        return bool(self._table)

    def __eq__(self, other: 'TableId') -> bool:
        return (  # ordering is for speed
            self._table == other._table and
            self._schema == other._schema and
            self._db == other._db
        )

    def __lt__(self, other: 'TableId') -> bool:
        return (
            (self._db, self._schema, self._table) <
            (other._db, other._schema, other._table)
        )

    def __hash__(self) -> int:
        return hash(str(self))

    def identifier(self, grammar: SqlGrammar) -> str:
        return make_identifier(grammar,
                               database=self._db,
                               schema=self._schema,
                               table=self._table)

    def schema_id(self) -> SchemaId:
        return SchemaId(db=self._db, schema=self._schema)

    def column_id(self, column: str) -> 'ColumnId':
        return ColumnId(db=self._db, schema=self._schema,
                        table=self._table, column=column)

    def database_schema_part(self, grammar: SqlGrammar) -> str:
        return make_identifier(grammar,
                               database=self._db,
                               schema=self._schema)

    def table_part(self, grammar: SqlGrammar) -> str:
        return make_identifier(grammar, table=self._table)

    def db(self) -> str:
        return self._db

    def schema(self) -> str:
        return self._schema

    def table(self) -> str:
        return self._table

    def __str__(self) -> str:
        return self.identifier(mysql_grammar)  # specific one unimportant

    def __repr__(self) -> str:
        return mapped_repr_stripping_underscores(
            self, ['_db', '_schema', '_table'])


@register_for_json(method=METHOD_STRIP_UNDERSCORE)
@functools.total_ordering
class ColumnId(object):
    def __init__(self, db: str = '', schema: str = '',
                 table: str = '', column: str = '') -> None:
        self._db = db
        self._schema = schema
        self._table = table
        self._column = column

    def __bool__(self) -> bool:
        return bool(self._column)

    def __eq__(self, other: 'ColumnId') -> bool:
        return (
            self._column == other._column and
            self._table == other._table and
            self._schema == other._schema and
            self._db == other._db
        )

    def __lt__(self, other: 'ColumnId') -> bool:
        return (
            (self._db, self._schema, self._table, self._column) <
            (other._db, other._schema, other._table, other._column)
        )

    def is_valid(self) -> bool:
        return bool(self._table and self._column)  # the minimum

    def identifier(self, grammar: SqlGrammar) -> str:
        return make_identifier(grammar,
                               database=self._db,
                               schema=self._schema,
                               table=self._table,
                               column=self._column)

    def db(self) -> str:
        return self._db

    def schema(self) -> str:
        return self._schema

    def table(self) -> str:
        return self._table

    def column(self) -> str:
        return self._column

    def schema_id(self) -> SchemaId:
        return SchemaId(db=self._db, schema=self._schema)

    def table_id(self) -> TableId:
        return TableId(db=self._db, schema=self._schema, table=self._table)

    def has_table_and_column(self) -> bool:
        return bool(self._table and self._column)

    def __str__(self) -> str:
        return self.identifier(mysql_grammar)  # specific one unimportant

    def __repr__(self) -> str:
        return mapped_repr_stripping_underscores(
            self, ['_db', '_schema', '_table', '_column'])

    # def html(self, grammar: SqlGrammar, bold_column: bool = True) -> str:
    #     components = [
    #         html.escape(grammar.quote_identifier_if_required(x))
    #         for x in [self._db, self._schema, self._table, self._column]
    #         if x]
    #     if not components:
    #         return ''
    #     if bold_column:
    #         components[-1] = "<b>{}</b>".format(components[-1])
    #     return ".".join(components)


def split_db_schema_table(db_schema_table: str) -> TableId:
    components = db_schema_table.split('.')
    if len(components) == 3:  # db.schema.table
        d, s, t = components[0], components[1], components[2]
    elif len(components) == 2:  # schema.table
        d, s, t = '', components[0], components[1]
    elif len(components) == 1:  # table
        d, s, t = '', '', components[0]
    else:
        raise ValueError("Bad db_schema_table: {}".format(db_schema_table))
    return TableId(db=d, schema=s, table=t)


def split_db_schema_table_column(db_schema_table_col: str) -> ColumnId:
    components = db_schema_table_col.split('.')
    if len(components) == 4:  # db.schema.table.column
        d, s, t, c = components[0], components[1], components[2], components[3]
    elif len(components) == 3:  # schema.table.column
        d, s, t, c = '', components[0], components[1], components[2]
    elif len(components) == 2:  # table.column
        d, s, t, c = '', '', components[0], components[1]
    elif len(components) == 1:  # column
        d, s, t, c = '', '', '', components[0]
    else:
        raise ValueError("Bad db_schema_table_col: {}".format(
            db_schema_table_col))
    return ColumnId(db=d, schema=s, table=t, column=c)


def columns_to_table_column_hierarchy(
        columns: List[ColumnId],
        sort: bool = True) -> List[Tuple[TableId, List[ColumnId]]]:
    tables = unique_list(c.table_id() for c in columns)
    if sort:
        tables.sort()
    table_column_map = []
    for t in tables:
        t_columns = [c for c in columns if c.table_id() == t]
        if sort:
            t_columns.sort()
        table_column_map.append((t, t_columns))
    return table_column_map


# =============================================================================
# Using SQL grammars (but without reference to Django models, for testing)
# =============================================================================

def make_identifier(grammar: SqlGrammar,
                    database: str = None,
                    schema: str = None,
                    table: str = None,
                    column: str = None) -> str:
    elements = [grammar.quote_identifier_if_required(x)
                for x in (database, schema, table, column) if x]
    assert elements, "make_identifier(): No elements passed!"
    return ".".join(elements)


def dumb_make_identifier(database: str = None,
                         schema: str = None,
                         table: str = None,
                         column: str = None) -> str:
    elements = filter(None, [database, schema, table, column])
    assert elements, "make_identifier(): No elements passed!"
    return ".".join(elements)


def parser_add_result_column(parsed: ParseResults,
                             column: str,
                             grammar: SqlGrammar) -> ParseResults:
    # Presupposes at least one column already in the SELECT statement.
    # log.critical("Adding: " + repr(column))
    existing_columns = parsed.select_expression.select_columns.asList()
    # log.critical(parsed.dump())
    # log.critical("existing columns: {}".format(repr(existing_columns)))
    # log.critical("adding column: {}".format(column))
    if column not in existing_columns:
        # log.critical("... doesn't exist; adding")
        newcol = grammar.get_result_column().parseString(column,
                                                         parseAll=True)
        # log.critical("... " + repr(newcol))
        parsed.select_expression.extend([",", newcol])
    # else:
    #     log.critical("... skipping column; exists")
    # log.critical(parsed.dump())
    return parsed


class JoinInfo(object):
    def __init__(self,
                 table: str,
                 join_type: str = 'INNER JOIN',
                 join_condition: str = '') -> None:  # e.g. "ON x = y"
        self.join_type = join_type
        self.table = table
        self.join_condition = join_condition


def parser_add_from_tables(parsed: ParseResults,
                           join_info_list: List[JoinInfo],
                           grammar: SqlGrammar) -> ParseResults:
    """
    joininfo: list of dictionaries with keys:
        table, join_type, join_condition
    Presupposes at least one table already in the FROM clause.
    """
    # log.critical(parsed.dump())
    existing_tables = parsed.join_source.from_tables.asList()
    # log.critical("existing tables: {}".format(existing_tables))
    # log.critical("adding table: {}".format(table))
    for ji in join_info_list:
        if ji.table in existing_tables:  # already there
            # log.critical("field already present")
            continue
        parsed_join = grammar.get_join_op().parseString(ji.join_type,
                                                        parseAll=True)[0]  # e.g. INNER JOIN  # noqa
        parsed_table = grammar.get_table_spec().parseString(ji.table,
                                                            parseAll=True)[0]
        extrabits = [parsed_join, parsed_table]
        if ji.join_condition:  # e.g. ON x = y
            extrabits.append(
                grammar.get_join_constraint().parseString(ji.join_condition,
                                                          parseAll=True)[0])
        parsed.join_source.extend(extrabits)
    # log.critical(parsed.dump())
    return parsed


def get_first_from_table(parsed: ParseResults,
                         match_db: str = '',
                         match_schema: str = '',
                         match_table: str = '') -> TableId:
    """
    Given a set of parsed results from a SELECT statement,
    returns the (db, schema, table) tuple
    representing the first table in the FROM clause.

    Optionally, the match may be constrained with the match* parameters.
    """
    existing_tables = parsed.join_source.from_tables.asList()
    for t in existing_tables:
        table_id = split_db_schema_table(t)
        if match_db and table_id.db() != match_db:
            continue
        if match_schema and table_id.schema() != match_schema:
            continue
        if match_table and table_id.table() != match_table:
            continue
        return table_id
    return TableId()


def set_distinct_within_parsed(p: ParseResults, action: str = 'set') -> None:
    ss = p.select_specifier  # type: ParseResults
    if action == 'set':
        if 'DISTINCT' not in ss.asList():
            ss.append('DISTINCT')
    elif action == 'clear':
        if 'DISTINCT' in ss.asList():
            del ss[:]
    elif action == 'toggle':
        if 'DISTINCT' in ss.asList():
            del ss[:]
        else:
            ss.append('DISTINCT')
    else:
        raise ValueError("action must be one of set/clear/toggle")


def set_distinct(sql: str,
                 grammar: SqlGrammar,
                 action: str = 'set',
                 formatted: bool = True,
                 debug: bool = False,
                 debug_verbose: bool = False) -> str:
    p = grammar.get_select_statement().parseString(sql, parseAll=True)
    if debug:
        log.info("START: {}".format(sql))
        if debug_verbose:
            log.debug("start dump:\n" + p.dump())
    set_distinct_within_parsed(p, action=action)
    result = text_from_parsed(p, formatted=formatted)
    if debug:
        log.info("END: {}".format(result))
        if debug_verbose:
            log.debug("end dump:\n" + p.dump())
    return result


def toggle_distinct(sql: str,
                    grammar: SqlGrammar,
                    formatted: bool = True,
                    debug: bool = False,
                    debug_verbose: bool = False) -> str:
    return set_distinct(sql=sql,
                        grammar=grammar,
                        action='toggle',
                        formatted=formatted,
                        debug=debug,
                        debug_verbose=debug_verbose)


# =============================================================================
# SQLAlchemy reflection and DDL
# =============================================================================

_print_not_execute = False


def set_print_not_execute(print_not_execute: bool) -> None:
    global _print_not_execute
    _print_not_execute = print_not_execute


def execute(engine: Engine, sql: str) -> None:
    log.debug(sql)
    if _print_not_execute:
        print(format_sql_for_print(sql) + "\n;")
        # extra \n in case the SQL ends in a comment
    else:
        engine.execute(sql)


def add_columns(engine: Engine, table: Table, columns: List[Column]) -> None:
    existing_column_names = get_column_names(engine, tablename=table.name,
                                             to_lower=True)
    column_defs = []
    for column in columns:
        if column.name.lower() not in existing_column_names:
            column_defs.append(column_creation_ddl(column, engine.dialect))
        else:
            log.debug("Table {}: column {} already exists; not adding".format(
                repr(table.name), repr(column.name)))
    # ANSI SQL: add one column at a time: ALTER TABLE ADD [COLUMN] coldef
    #   - i.e. "COLUMN" optional, one at a time, no parentheses
    #   - http://www.contrib.andrew.cmu.edu/~shadow/sql/sql1992.txt
    # MySQL: ALTER TABLE ADD [COLUMN] (a INT, b VARCHAR(32));
    #   - i.e. "COLUMN" optional, parentheses required for >1, multiple OK
    #   - http://dev.mysql.com/doc/refman/5.7/en/alter-table.html
    # MS SQL Server: ALTER TABLE ADD COLUMN a INT, B VARCHAR(32);
    #   - i.e. no "COLUMN", no parentheses, multiple OK
    #   - https://msdn.microsoft.com/en-us/library/ms190238.aspx
    #   - https://msdn.microsoft.com/en-us/library/ms190273.aspx
    #   - http://stackoverflow.com/questions/2523676
    # SQLAlchemy doesn't provide a shortcut for this.
    for column_def in column_defs:
        log.info("Table {}: adding column {}".format(
            repr(table.name), repr(column_def)))
        execute(engine, """
            ALTER TABLE {tablename} ADD {column_def}
        """.format(tablename=table.name, column_def=column_def))


def drop_columns(engine: Engine, table: Table,
                 column_names: Iterable[str]) -> None:
    existing_column_names = get_column_names(engine, tablename=table.name,
                                             to_lower=True)
    for name in column_names:
        if name.lower() not in existing_column_names:
            log.debug("Table {}: column {} does not exist; not "
                      "dropping".format(repr(table.name), repr(name)))
        else:
            log.info("Table {}: dropping column {}".format(
                repr(table.name), repr(name)))
            sql = "ALTER TABLE {t} DROP COLUMN {c}".format(t=table.name,
                                                           c=name)
            # SQL Server:
            #   http://www.techonthenet.com/sql_server/tables/alter_table.php
            # MySQL:
            #   http://dev.mysql.com/doc/refman/5.7/en/alter-table.html
            execute(engine, sql)


def add_indexes(engine: Engine, table: Table,
                indexdictlist: Iterable[Dict[str, Any]]) -> None:
    existing_index_names = get_index_names(engine, tablename=table.name,
                                           to_lower=True)
    for idxdefdict in indexdictlist:
        index_name = idxdefdict['index_name']
        column = idxdefdict['column']
        if not isinstance(column, str):
            column = ", ".join(column)  # must be a list
        unique = idxdefdict.get('unique', False)
        if index_name.lower() not in existing_index_names:
            log.info("Table {}: adding index {} on column {}".format(
                repr(table.name), repr(index_name), repr(column)))
            execute(engine, """
                CREATE{unique} INDEX {idxname} ON {tablename} ({column})
            """.format(
                unique=" UNIQUE" if unique else "",
                idxname=index_name,
                tablename=table.name,
                column=column,
            ))
        else:
            log.debug("Table {}: index {} already exists; not adding".format(
                repr(table.name), repr(index_name)))


def drop_indexes(engine: Engine, table: Table,
                 index_names: Iterable[str]) -> None:
    existing_index_names = get_index_names(engine, tablename=table.name,
                                           to_lower=True)
    for index_name in index_names:
        if index_name.lower() not in existing_index_names:
            log.debug("Table {}: index {} does not exist; not dropping".format(
                repr(table.name), repr(index_name)))
        else:
            log.info("Table {}: dropping index {}".format(
                repr(table.name), repr(index_name)))
            if engine.dialect.name == 'mysql':
                sql = "ALTER TABLE {t} DROP INDEX {i}".format(t=table.name,
                                                              i=index_name)
            elif engine.dialect.name == 'mssql':
                sql = "DROP INDEX {t}.{i}".format(t=table.name, i=index_name)
            else:
                assert False, "Unknown dialect: {}".format(engine.dialect.name)
            execute(engine, sql)


def get_table_names(engine: Engine,
                    to_lower: bool = False,
                    sort: bool = False) -> List[str]:
    inspector = inspect(engine)
    table_names = inspector.get_table_names()
    if to_lower:
        table_names = [x.lower() for x in table_names]
    if sort:
        table_names = sorted(table_names, key=lambda x: x.lower())
    return table_names


def get_view_names(engine: Engine,
                   to_lower: bool = False,
                   sort: bool = False) -> List[str]:
    inspector = inspect(engine)
    view_names = inspector.get_view_names()
    if to_lower:
        view_names = [x.lower() for x in view_names]
    if sort:
        view_names = sorted(view_names, key=lambda x: x.lower())
    return view_names


def get_column_names(engine: Engine,
                     tablename: str,
                     to_lower: bool = False,
                     sort: bool = False) -> List[str]:
    """
    Reads columns names afresh from the database (in case metadata is out of
    date).
    """
    inspector = inspect(engine)
    columns = inspector.get_columns(tablename)
    column_names = [x['name'] for x in columns]
    if to_lower:
        column_names = [x.lower() for x in column_names]
    if sort:
        column_names = sorted(column_names, key=lambda x: x.lower())
    return column_names


def get_index_names(engine: Engine,
                    tablename: str,
                    to_lower: bool = False) -> List[str]:
    """
    Reads index names from the database.
    """
    # http://docs.sqlalchemy.org/en/latest/core/reflection.html
    inspector = inspect(engine)
    indexes = inspector.get_indexes(tablename)
    index_names = [x['name'] for x in indexes if x['name']]
    # ... at least for SQL Server, there always seems to be a blank one
    # with {'name': None, ...}.
    if to_lower:
        index_names = [x.lower() for x in index_names]
    return index_names


def ensure_columns_present(engine: Engine,
                           tablename: str,
                           column_names: Iterable[str]) -> None:
    existing_column_names = get_column_names(engine, tablename=tablename,
                                             to_lower=True)
    if not column_names:
        return
    for col in column_names:
        if col.lower() not in existing_column_names:
            raise ValueError(
                "Column {} missing from table {}".format(
                    repr(col), repr(tablename)))


def create_view(engine: Engine,
                viewname: str,
                select_sql: str) -> None:
    if engine.dialect.name == 'mysql':
        # MySQL has CREATE OR REPLACE VIEW.
        sql = "CREATE OR REPLACE VIEW {viewname} AS {select_sql}".format(
            viewname=viewname,
            select_sql=select_sql,
        )
    else:
        # SQL Server doesn't: http://stackoverflow.com/questions/18534919
        drop_view(engine, viewname, quiet=True)
        sql = "CREATE VIEW {viewname} AS {select_sql}".format(
            viewname=viewname,
            select_sql=select_sql,
        )
    log.info("Creating view: {}".format(repr(viewname)))
    execute(engine, sql)


def assert_view_has_same_num_rows(engine: Engine,
                                  basetable: str,
                                  viewname: str) -> None:
    # Note that this relies on the data, i.e. design failures MAY cause this
    # assertion to fail, but won't necessarily (e.g. if the table is empty).
    n_base = count_star(engine, basetable)
    n_view = count_star(engine, viewname)
    assert n_view == n_base, (
        "View bug: view {} has {} records but its base table {} "
        "has {}; they should be equal".format(
            viewname, n_view,
            basetable, n_base))


def drop_view(engine: Engine,
              viewname: str,
              quiet: bool = False) -> None:
    # MySQL has DROP VIEW IF EXISTS, but SQL Server only has that from
    # SQL Server 2016 onwards.
    # - https://msdn.microsoft.com/en-us/library/ms173492.aspx
    # - http://dev.mysql.com/doc/refman/5.7/en/drop-view.html
    view_names = get_view_names(engine, to_lower=True)
    if viewname.lower() not in view_names:
        log.debug("View {} does not exist; not dropping".format(viewname))
    else:
        if not quiet:
            log.info("Dropping view: {}".format(repr(viewname)))
        sql = "DROP VIEW {viewname}".format(viewname=viewname)
        execute(engine, sql)


# =============================================================================
# View-building assistance class
# =============================================================================

class ViewMaker(object):
    def __init__(self,
                 viewname: str,
                 engine: Engine,
                 basetable: str,
                 existing_to_lower: bool = False,
                 rename: Dict[str, str] = None,
                 progargs: argparse.Namespace = None,
                 enforce_same_n_rows_as_base: bool = True) -> None:
        rename = rename or {}
        self.viewname = viewname
        self.engine = engine
        self.basetable = basetable
        self.progargs = progargs  # only for others' benefit
        self.enforce_same_n_rows_as_base = enforce_same_n_rows_as_base

        self.select_elements = []
        for colname in get_column_names(engine, tablename=basetable,
                                        to_lower=existing_to_lower):
            if colname in rename:
                rename_to = rename[colname]
                if not rename_to:
                    continue
                as_clause = " AS {}".format(rename_to)
            else:
                as_clause = ""
            self.select_elements.append("{t}.{c}{as_clause}".format(
                t=basetable, c=colname, as_clause=as_clause))
        assert self.select_elements, "Must have some active SELECT elements " \
                                     "from base table"
        self.from_elements = [basetable]
        self.where_elements = []
        self.lookup_table_keyfields = []  # of (table, keyfield(s)) tuples

    def add_select(self, clause: str) -> None:
        self.select_elements.append(clause)

    def add_from(self, clause: str) -> None:
        self.from_elements.append(clause)

    def add_where(self, clause: str) -> None:
        self.where_elements.append(clause)

    def get_sql(self) -> str:
        if self.where_elements:
            where = "\n    WHERE {}".format(
                "\n        AND ".join(self.where_elements))
        else:
            where = ""
        return (
            "\n    SELECT {select_elements}"
            "\n    FROM {from_elements}{where}".format(
                select_elements=",\n        ".join(self.select_elements),
                from_elements="\n        ".join(self.from_elements),
                where=where))

    def create_view(self, engine: Engine) -> None:
        create_view(engine, self.viewname, self.get_sql())
        if self.enforce_same_n_rows_as_base:
            assert_view_has_same_num_rows(engine, self.basetable,
                                          self.viewname)

    def drop_view(self, engine: Engine) -> None:
        drop_view(engine, self.viewname)

    def record_lookup_table_keyfield(
            self,
            table: str,
            keyfield: Union[str, Iterable[str]]) -> None:
        self.lookup_table_keyfields.append((table, keyfield))

    def record_lookup_table_keyfields(
            self,
            table_keyfield_tuples: Iterable[
                Tuple[str, Union[str, Iterable[str]]]
            ]) -> None:
        for t, k in table_keyfield_tuples:
            self.record_lookup_table_keyfield(t, k)

    def get_lookup_tables(self) -> List[str]:
        return list(set(table for table, keyfield
                        in self.lookup_table_keyfields))

    def get_lookup_table_keyfields(self) -> List[Tuple[str, str]]:
        return list(self.lookup_table_keyfields)


# =============================================================================
# Transaction size-limiting class
# =============================================================================

class TransactionSizeLimiter(object):
    def __init__(self, session: Session,
                 max_rows_before_commit: int = None,
                 max_bytes_before_commit: int = None) -> None:
        self._session = session
        self._max_rows_before_commit = max_rows_before_commit
        self._max_bytes_before_commit = max_bytes_before_commit
        self._bytes_in_transaction = 0
        self._rows_in_transaction = 0

    def commit(self) -> None:
        with MultiTimerContext(timer, TIMING_COMMIT):
            self._session.commit()
        self._bytes_in_transaction = 0
        self._rows_in_transaction = 0

    def notify(self, n_rows: int, n_bytes: int,
               force_commit: bool=False) -> None:
        if force_commit:
            self.commit()
            return
        self._bytes_in_transaction += n_bytes
        self._rows_in_transaction += n_rows
        # log.critical(
        #     "adding {} rows, {} bytes, "
        #     "to make {} rows, {} bytes so far".format(
        #         n_rows, n_bytes,
        #         self._rows_in_transaction, self._bytes_in_transaction))
        if (self._max_bytes_before_commit is not None and
                self._bytes_in_transaction >= self._max_bytes_before_commit):
            log.info(
                "Triggering early commit based on byte count (reached {}, "
                "limit is {})".format(
                    sizeof_fmt(self._bytes_in_transaction),
                    sizeof_fmt(self._max_bytes_before_commit)))
            self.commit()
        elif (self._max_rows_before_commit is not None and
                self._rows_in_transaction >= self._max_rows_before_commit):
            log.info(
                "Triggering early commit based on row count (reached {} rows, "
                "limit is {})".format(self._rows_in_transaction,
                                      self._max_rows_before_commit))
            self.commit()


# =============================================================================
# Specification matching
# =============================================================================

def _matches_tabledef(table: str, tabledef: str) -> bool:
    tr = get_spec_match_regex(tabledef)
    return tr.match(table)


def matches_tabledef(table: str, tabledef: Union[str, List[str]]) -> bool:
    if isinstance(tabledef, str):
        return _matches_tabledef(table, tabledef)
    elif not tabledef:
        return False
    else:  # list
        return any(_matches_tabledef(table, td) for td in tabledef)


def _matches_fielddef(table: str, field: str, fielddef: str) -> bool:
    column_id = split_db_schema_table_column(fielddef)
    cr = get_spec_match_regex(column_id.column())
    if not column_id.table():
        return cr.match(field)
    tr = get_spec_match_regex(column_id.table())
    return tr.match(table) and cr.match(field)


def matches_fielddef(table: str, field: str,
                     fielddef: Union[str, List[str]]) -> bool:
    if isinstance(fielddef, str):
        return _matches_fielddef(table, field, fielddef)
    elif not fielddef:
        return False
    else:  # list
        return any(_matches_fielddef(table, field, fd) for fd in fielddef)


# =============================================================================
# More SQL
# =============================================================================

def sql_fragment_cast_to_int(expr: str,
                             big: bool = True,
                             dialect: Dialect = None,
                             viewmaker: ViewMaker = None) -> str:
    """
    For Microsoft SQL Server.
    Conversion to INT:
    - http://stackoverflow.com/questions/2000045
    - http://stackoverflow.com/questions/14719760  # this one
    - http://stackoverflow.com/questions/14692131
      ... LIKE example.
      ... ISNUMERIC()
          https://msdn.microsoft.com/en-us/library/ms186272.aspx
          ... but that includes non-integer numerics
    - https://msdn.microsoft.com/en-us/library/ms174214(v=sql.120).aspx
      ... relates to the SQL Server Management Studio "Find and Replace"
          dialogue box, not to SQL itself!
    - http://stackoverflow.com/questions/29206404/mssql-regular-expression

    Note that the regex-like expression supported by LIKE is extremely limited.
    - https://msdn.microsoft.com/en-us/library/ms179859.aspx
    The only things supported are:

        %   any characters
        _   any single character
        []  single character in range or set, e.g. [a-f], [abcdef]
        [^] single character NOT in range or set, e.g. [^a-f], [abcdef]

    SQL Server does not support a REGEXP command directly.

    So the best bet is to have the LIKE clause check for a non-integer:

        CASE
            WHEN something LIKE '%[^0-9]%' THEN NULL
            ELSE CAST(something AS BIGINT)
        END

    ... which doesn't deal with spaces properly, but there you go.
    Could also strip whitespace left/right:

        CASE
            WHEN LTRIM(RTRIM(something)) LIKE '%[^0-9]%' THEN NULL
            ELSE CAST(something AS BIGINT)
        END

    Only works for positive integers.
    LTRIM/RTRIM are not ANSI SQL.
    Nor are unusual LIKE clauses; see
        http://stackoverflow.com/questions/712580/list-of-special-characters-for-sql-like-clause

    The other, for SQL Server 2012 or higher, is TRY_CAST:

        TRY_CAST(something AS BIGINT)
        ... returns NULL upon failure
        ... https://msdn.microsoft.com/en-us/library/hh974669.aspx
    """  # noqa
    inttype = "BIGINT" if big else "INTEGER"
    if dialect is None and viewmaker is not None:
        dialect = viewmaker.engine.dialect
    if dialect is None:
        sql_server = True
        supports_try_cast = False
    else:
        # noinspection PyUnresolvedReferences
        sql_server = dialect.name == 'mssql'
        # noinspection PyUnresolvedReferences
        supports_try_cast = (sql_server and
                             dialect.server_version_info >= MS_2012_VERSION)
    if supports_try_cast:
        return "TRY_CAST({expr} AS {inttype})".format(expr=expr,
                                                      inttype=inttype)
    elif sql_server:
        return (
            "CASE WHEN LTRIM(RTRIM({expr})) LIKE '%[^0-9]%' "
            "THEN NULL ELSE CAST({expr} AS {inttype}) END".format(
                expr=expr, inttype=inttype)
        )
        # Doesn't support negative integers.
    else:
        # noinspection PyUnresolvedReferences
        raise ValueError("Code not yet written for convert-to-int for "
                         "dialect {}".format(dialect.name))


# =============================================================================
# Abstracted SQL WHERE condition
# =============================================================================

@register_for_json(method=METHOD_PROVIDES_INIT_KWARGS)
@functools.total_ordering
class WhereCondition(object):
    # Ancillary class for building SQL WHERE expressions from our web forms.
    def __init__(self,
                 column_id: ColumnId = None,
                 op: str = '',
                 datatype: str = '',
                 value_or_values: Any = None,
                 raw_sql: str = '',
                 from_table_for_raw_sql: TableId = None) -> None:
        self._column_id = column_id
        self._op = op.upper()
        self._datatype = datatype
        self._value = value_or_values
        self._no_value = False
        self._multivalue = False
        self._raw_sql = raw_sql
        self._from_table_for_raw_sql = from_table_for_raw_sql

        if not self._raw_sql:
            if self._op in SQL_OPS_VALUE_UNNECESSARY:
                self._no_value = True
                assert value_or_values is None, "Superfluous value passed"
            elif self._op in SQL_OPS_MULTIPLE_VALUES:
                self._multivalue = True
                assert isinstance(value_or_values, list), "Need list"
            else:
                assert not isinstance(value_or_values, list), "Need single value"  # noqa

    def init_kwargs(self) -> Dict:
        return {
            'column_id': self._column_id,
            'op': self._op,
            'datatype': self._datatype,
            'value_or_values': self._value,
            'raw_sql': self._raw_sql,
            'from_table_for_raw_sql': self._from_table_for_raw_sql,
        }

    def __repr__(self) -> str:
        return (
            "<{qualname}("
            "column_id={column_id}, "
            "op={op}, "
            "datatype={datatype}, "
            "value_or_values={value_or_values}, "
            "raw_sql={raw_sql}, "
            "from_table_for_raw_sql={from_table_for_raw_sql}"
            ") at {addr}>".format(
                qualname=self.__class__.__qualname__,
                column_id=repr(self._column_id),
                op=repr(self._op),
                datatype=repr(self._datatype),
                value_or_values=repr(self._value),
                raw_sql=repr(self._raw_sql),
                from_table_for_raw_sql=repr(self._from_table_for_raw_sql),
                addr=hex(id(self)),
            )
        )

    def __eq__(self, other: 'WhereCondition') -> bool:
        return (
            self._raw_sql == other._raw_sql and
            self._column_id == other._column_id and
            self._op == other._op and
            self._value == other._value
        )

    def __lt__(self, other: 'WhereCondition') -> bool:
        return (
            (self._raw_sql, self._column_id, self._op, self._value) <
            (other._raw_sql, other._column_id, other._op, other._value)
        )

    def column_id(self) -> ColumnId:
        return self._column_id

    def table_id(self) -> TableId:
        if self._raw_sql:
            return self._from_table_for_raw_sql
        return self.column_id().table_id()

    def table_str(self, grammar: SqlGrammar) -> str:
        return self.table_id().identifier(grammar)

    def sql(self, grammar: SqlGrammar) -> str:
        if self._raw_sql:
            return self._raw_sql

        col = self._column_id.identifier(grammar)
        op = self._op

        if self._no_value:
            return "{col} {op}".format(col=col, op=op)

        if self._datatype in QB_STRING_TYPES:
            element_converter = sql_string_literal
        elif self._datatype == QB_DATATYPE_DATE:
            element_converter = sql_date_literal
        elif self._datatype == QB_DATATYPE_INTEGER:
            element_converter = str
        elif self._datatype == QB_DATATYPE_FLOAT:
            element_converter = str
        else:
            # Safe default
            element_converter = sql_string_literal

        if self._multivalue:
            literal = "({})".format(", ".join(element_converter(v)
                                              for v in self._value))
        else:
            literal = element_converter(self._value)

        if self._op == 'MATCH':  # MySQL
            return "MATCH ({col}) AGAINST ({val})".format(col=col, val=literal)
        elif self._op == 'CONTAINS':  # SQL Server
            return "CONTAINS({col}, {val})".format(col=col, val=literal)
        else:
            return "{col} {op} {val}".format(col=col, op=op, val=literal)


# =============================================================================
# SQL formatting
# =============================================================================

def format_sql_for_print(sql: str) -> str:
    # Remove blank lines and trailing spaces
    lines = list(filter(None, [x.replace("\t", "    ").rstrip()
                               for x in sql.splitlines()]))
    # Shift all lines left if they're left-padded
    firstleftpos = float('inf')
    for line in lines:
        leftpos = len(line) - len(line.lstrip())
        firstleftpos = min(firstleftpos, leftpos)
    if firstleftpos > 0:
        lines = [x[firstleftpos:] for x in lines]
    return "\n".join(lines)


# =============================================================================
# Plain SQL types
# =============================================================================

def is_sql_column_type_textual(column_type: str,
                               min_length: int = 1) -> bool:
    column_type = column_type.upper()
    if column_type == 'TEXT':
        return True
    try:
        m = COLTYPE_WITH_ONE_INTEGER_REGEX.match(column_type)
        basetype = m.group(1)
        length = int(m.group(2))
    except (AttributeError, ValueError):
        return False
    return length >= min_length and basetype in SQLTYPES_TEXT


def escape_quote_in_literal(s: str) -> str:
    """
    Escape '. We could use '' or \'.
    Let's use \. for consistency with percent escaping.
    """
    return s.replace("'", r"\'")


def escape_percent_in_literal(sql: str) -> str:
    """
    Escapes % by converting it to \%.
    Use this for LIKE clauses.
    http://dev.mysql.com/doc/refman/5.7/en/string-literals.html
    """
    return sql.replace('%', r'\%')


def escape_percent_for_python_dbapi(sql: str) -> str:
    """
    Escapes % by converting it to %%.
    Use this for SQL within Python where % characters are used for argument
    placeholders.
    """
    return sql.replace('%', '%%')


def escape_sql_string_literal(s: str) -> str:
    """
    Escapes SQL string literal fragments against quotes and parameter
    substitution.
    """
    return escape_percent_in_literal(escape_quote_in_literal(s))


def translate_sql_qmark_to_percent(sql: str) -> str:
    """
    MySQL likes '?' as a placeholder.
    - https://dev.mysql.com/doc/refman/5.7/en/sql-syntax-prepared-statements.html

    Python DBAPI allows several: '%s', '?', ':1', ':name', '%(name)s'.
    - https://www.python.org/dev/peps/pep-0249/#paramstyle

    Django uses '%s'.
    - https://docs.djangoproject.com/en/1.8/topics/db/sql/

    Microsoft like '?', '@paramname', and ':paramname'.
    - https://msdn.microsoft.com/en-us/library/yy6y35y8(v=vs.110).aspx

    We need to parse SQL with argument placeholders.
    - See SqlGrammar classes, particularly: bind_parameter

    I prefer ?, because % is used in LIKE clauses, and the databases we're
    using like it.

    So:
    - We use %s when using cursor.execute() directly, via Django.
    - We use ? when talking to users, and SqlGrammar objects, so that the
      visual appearance matches what they expect from their database.

    This function translates SQL using ? placeholders to SQL using %s
    placeholders, without breaking literal '?' or '%', e.g. inside string
    literals.
    """  # noqa
    # 1. Escape % characters
    sql = escape_percent_for_python_dbapi(sql)
    # 2. Replace ? characters that are not within quotes with %s.
    newsql = ""
    in_quotes = False
    for c in sql:
        if c == "'":
            in_quotes = not in_quotes
        if c == '?' and not in_quotes:
            newsql += '%s'
        else:
            newsql += c
    return newsql


if False:
    _SQLTEST1 = "SELECT a FROM b WHERE c=? AND d LIKE 'blah%' AND e='?'"
    _SQLTEST2 = "SELECT a FROM b WHERE c=%s AND d LIKE 'blah%%' AND e='?'"
    _SQLTEST3 = translate_sql_qmark_to_percent(_SQLTEST1)


# =============================================================================
# Tests
# =============================================================================

def unit_tests():
    assert matches_tabledef("sometable", "sometable")
    assert matches_tabledef("sometable", "some*")
    assert matches_tabledef("sometable", "*table")
    assert matches_tabledef("sometable", "*")
    assert matches_tabledef("sometable", "s*e")
    assert not matches_tabledef("sometable", "x*y")

    assert matches_fielddef("sometable", "somefield", "*.somefield")
    assert matches_fielddef("sometable", "somefield", "sometable.somefield")
    assert matches_fielddef("sometable", "somefield", "sometable.*")
    assert matches_fielddef("sometable", "somefield", "somefield")

    grammar = make_grammar(DIALECT_MYSQL)
    sql = "SELECT t1.c1, t2.c2 " \
          "FROM t1 INNER JOIN t2 ON t1.k = t2.k"
    parsed = grammar.get_select_statement().parseString(sql, parseAll=True)
    table_id = get_first_from_table(parsed)  # noqa
    log.info(repr(table_id))


if __name__ == '__main__':
    main_only_quicksetup_rootlogger()
    unit_tests()



#!/usr/bin/env python
# crate_anon/crateweb/research/sql_grammar_mysql.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
import re

from pyparsing import (
    Combine,
    cStyleComment,
    delimitedList,
    Forward,
    Group,
    infixNotation,
    Literal,
    NotAny,
    oneOf,
    OneOrMore,
    opAssoc,
    Optional,
    QuotedString,
    Regex,
    ZeroOrMore,
)

from crate_anon.common.logsupport import main_only_quicksetup_rootlogger
from crate_anon.common.sql_grammar import (
    ALL,
    AND,
    ansi_comment,
    ANSI92_RESERVED_WORD_LIST,
    AS,
    ASC,
    AVG,
    bash_comment,
    BETWEEN,
    BY,
    CASE,
    COLLATE,
    COMMA,
    CROSS,
    COUNT,
    delim_list,
    DESC,
    DISTINCT,
    ELSE,
    END,
    EXISTS,
    FOR,
    FROM,
    GROUP,
    HAVING,
    IN,
    INDEX,
    INNER,
    integer,
    INTERVAL,
    IS,
    JOIN,
    KEY,
    LANGUAGE,
    LEFT,
    LIKE,
    literal_value,
    LPAR,
    make_pyparsing_regex,
    make_regex_except_words,
    make_words_regex,
    MATCH,
    MAX,
    MIN,
    NATURAL,
    NOT,
    ON,
    OR,
    ORDER,
    OUTER,
    RIGHT,
    RPAR,
    SELECT,
    string_literal,
    SUM,
    sql_keyword,
    SqlGrammar,
    test_fail,
    test_succeed,
    THEN,
    time_unit,
    UNION,
    USE,
    USING,
    WHEN,
    WHERE,
    WITH,
)

log = logging.getLogger(__name__)

AGAINST = sql_keyword("AGAINST")
BIT_AND = sql_keyword("BIT_AND")
BIT_OR = sql_keyword("BIT_OR")
BIT_XOR = sql_keyword("BIT_XOR")
BOOLEAN = sql_keyword("BOOLEAN")
EXPANSION = sql_keyword("EXPANSION")
GROUP_CONCAT = sql_keyword("GROUP_CONCAT")
MODE = sql_keyword("MODE")
NULLS = sql_keyword("NULLS")
QUERY = sql_keyword("QUERY")
ROW = sql_keyword("ROW")
TABLESPACE = sql_keyword("TABLESPACE")
BINARY = sql_keyword("BINARY")
DISTINCTROW = sql_keyword("DISTINCTROW")
DIV = sql_keyword("DIV")
FORCE = sql_keyword("FORCE")
HIGH_PRIORITY = sql_keyword("HIGH_PRIORITY")
IGNORE = sql_keyword("IGNORE")
LIMIT = sql_keyword("LIMIT")
MAX_STATEMENT_TIME = sql_keyword("MAX_STATEMENT_TIME")
MOD = sql_keyword("MOD")
OFFSET = sql_keyword("OFFSET")
OJ = sql_keyword("OJ")
PARTITION = sql_keyword("PARTITION")
PROCEDURE = sql_keyword("PROCEDURE")
REGEXP = sql_keyword("REGEXP")
ROLLUP = sql_keyword("ROLLUP")
SOUNDS = sql_keyword("SOUNDS")
STD = sql_keyword("STD")
STDDEV = sql_keyword("STDDEV")
STDDEV_POP = sql_keyword("STDDEV_POP")
STDDEV_SAMP = sql_keyword("STDDEV_SAMP")
SQL_BIG_RESULT = sql_keyword("SQL_BIG_RESULT")
SQL_BUFFER_RESULT = sql_keyword("SQL_BUFFER_RESULT")
SQL_CACHE = sql_keyword("SQL_CACHE")
SQL_CALC_FOUND_ROWS = sql_keyword("SQL_CALC_FOUND_ROWS")
SQL_NO_CACHE = sql_keyword("SQL_NO_CACHE")
SQL_SMALL_RESULT = sql_keyword("SQL_SMALL_RESULT")
STRAIGHT_JOIN = sql_keyword("STRAIGHT_JOIN")
VAR_POP = sql_keyword("VAR_POP")
VAR_SAMP = sql_keyword("VAR_SAMP")
VARIANCE = sql_keyword("VARIANCE")
XOR = sql_keyword("XOR")


# =============================================================================
# MySQL 5.7 grammar in pyparsing
# =============================================================================
# http://dev.mysql.com/doc/refman/5.7/en/dynindex-statement.html

class SqlGrammarMySQL(SqlGrammar):
    # -------------------------------------------------------------------------
    # Forward declarations
    # -------------------------------------------------------------------------
    expr = Forward()
    select_statement = Forward()

    # -------------------------------------------------------------------------
    # Keywords
    # -------------------------------------------------------------------------
    # https://dev.mysql.com/doc/refman/5.7/en/keywords.html
    mysql_reserved_words = """
ACCESSIBLE ADD ALL ALTER ANALYZE AND AS ASC ASENSITIVE
BEFORE BETWEEN BIGINT BINARY BLOB BOTH BY
CALL CASCADE CASE CHANGE CHAR CHARACTER CHECK COLLATE COLUMN CONDITION
    CONSTRAINT CONTINUE CONVERT CREATE CROSS CURRENT_DATE CURRENT_TIME
    CURRENT_TIMESTAMP CURRENT_USER CURSOR
DATABASE DATABASES DAY_HOUR DAY_MICROSECOND DAY_MINUTE DAY_SECOND DEC DECIMAL
    DECLARE DEFAULT DELAYED DELETE DESC DESCRIBE DETERMINISTIC DISTINCT
    DISTINCTROW DIV DOUBLE DROP DUAL
EACH ELSE ELSEIF ENCLOSED ESCAPED EXISTS EXIT EXPLAIN
FALSE FETCH FLOAT FLOAT4 FLOAT8 FOR FORCE FOREIGN FROM FULLTEXT
GENERATED GET GRANT GROUP
HAVING HIGH_PRIORITY HOUR_MICROSECOND HOUR_MINUTE HOUR_SECOND
IF IGNORE IN INDEX INFILE INNER INOUT INSENSITIVE INSERT INT INT1 INT2 INT3
    INT4 INT8 INTEGER INTERVAL INTO IO_AFTER_GTIDS IO_BEFORE_GTIDS IS
    ITERATE
JOIN
KEY KEYS KILL
LEADING LEAVE LEFT LIKE LIMIT LINEAR LINES LOAD LOCALTIME LOCALTIMESTAMP
    LOCK LONG LONGBLOB LONGTEXT LOOP LOW_PRIORITY
MASTER_BIND MASTER_SSL_VERIFY_SERVER_CERT MATCH MAXVALUE MEDIUMBLOB
    MEDIUMINT MEDIUMTEXT MIDDLEINT MINUTE_MICROSECOND MINUTE_SECOND MOD
    MODIFIES
NATURAL NOT NO_WRITE_TO_BINLOG NULL NUMERIC
ON OPTIMIZE OPTIMIZER_COSTS OPTION OPTIONALLY OR ORDER OUT OUTER OUTFILE
PARTITION PRECISION PRIMARY PROCEDURE PURGE
RANGE READ READS READ_WRITE REAL REFERENCES REGEXP RELEASE RENAME REPEAT
    REPLACE REQUIRE RESIGNAL RESTRICT RETURN REVOKE RIGHT RLIKE
SCHEMA SCHEMAS SECOND_MICROSECOND SELECT SENSITIVE SEPARATOR SET SHOW SIGNAL
    SMALLINT SPATIAL SPECIFIC SQL SQLEXCEPTION SQLSTATE SQLWARNING
    SQL_BIG_RESULT SQL_CALC_FOUND_ROWS SQL_SMALL_RESULT SSL STARTING STORED
    STRAIGHT_JOIN
TABLE TERMINATED THEN TINYBLOB TINYINT TINYTEXT TO TRAILING TRIGGER TRUE
UNDO UNION UNIQUE UNLOCK UNSIGNED UPDATE USAGE USE USING UTC_DATE UTC_TIME
    UTC_TIMESTAMP
VALUES VARBINARY VARCHAR VARCHARACTER VARYING VIRTUAL
WHEN WHERE WHILE WITH WRITE
XOR
YEAR_MONTH
ZEROFILL
    """
    mysql_nonreserved_keywords = """
ACCOUNT ACTION AFTER AGAINST AGGREGATE ALGORITHM ALWAYS ANALYSE ANY ASCII AT
    AUTOEXTEND_SIZE AUTO_INCREMENT AVG AVG_ROW_LENGTH
BACKUP BEGIN BINLOG BIT BLOCK BOOL BOOLEAN BTREE BYTE
CACHE CASCADED CATALOG_NAME CHAIN CHANGED CHANNEL CHARSET CHECKSUM CIPHER
    CLASS_ORIGIN CLIENT CLOSE COALESCE CODE COLLATION COLUMNS COLUMN_FORMAT
    COLUMN_NAME COMMENT COMMIT COMMITTED COMPACT COMPLETION COMPRESSED
    COMPRESSION CONCURRENT CONNECTION CONSISTENT CONSTRAINT_CATALOG
    CONSTRAINT_NAME CONSTRAINT_SCHEMA CONTAINS CONTEXT CPU CUBE CURRENT
    CURSOR_NAME
DATA DATAFILE DATE DATETIME DAY DEALLOCATE DEFAULT_AUTH DEFINER DELAY_KEY_WRITE
    DES_KEY_FILE DIAGNOSTICS DIRECTORY DISABLE DISCARD DISK DO DUMPFILE
    DUPLICATE DYNAMIC
ENABLE ENCRYPTION END ENDS ENGINE ENGINES ENUM ERROR ERRORS ESCAPE EVENT EVENTS
    EVERY EXCHANGE EXECUTE EXPANSION EXPIRE EXPORT EXTENDED EXTENT_SIZE
FAST FAULTS FIELDS FILE FILE_BLOCK_SIZE FILTER FIRST FIXED FLUSH FOLLOWS FORMAT
    FOUND FULL FUNCTION
GENERAL GEOMETRY GEOMETRYCOLLECTION GET_FORMAT GLOBAL GRANTS GROUP_REPLICATION
HANDLER HASH HELP HOST HOSTS HOUR
IDENTIFIED IGNORE_SERVER_IDS IMPORT INDEXES INITIAL_SIZE INSERT_METHOD INSTALL
    INSTANCE INVOKER IO IO_THREAD IPC ISOLATION ISSUER
JSON
KEY_BLOCK_SIZE
LANGUAGE LAST LEAVES LESS LEVEL LINESTRING LIST LOCAL LOCKS LOGFILE LOGS
MASTER MASTER_AUTO_POSITION MASTER_CONNECT_RETRY MASTER_DELAY
    MASTER_HEARTBEAT_PERIOD MASTER_HOST MASTER_LOG_FILE MASTER_LOG_POS
    MASTER_PASSWORD MASTER_PORT MASTER_RETRY_COUNT MASTER_SERVER_ID MASTER_SSL
    MASTER_SSL_CA MASTER_SSL_CAPATH MASTER_SSL_CERT MASTER_SSL_CIPHER
    MASTER_SSL_CRL MASTER_SSL_CRLPATH MASTER_SSL_KEY MASTER_TLS_VERSION
    MASTER_USER MAX_CONNECTIONS_PER_HOUR MAX_QUERIES_PER_HOUR MAX_ROWS MAX_SIZE
    MAX_STATEMENT_TIME MAX_UPDATES_PER_HOUR MAX_USER_CONNECTIONS MEDIUM
    MEMORY MERGE MESSAGE_TEXT MICROSECOND MIGRATE MINUTE MIN_ROWS MODE MODIFY
    MONTH MULTILINESTRING MULTIPOINT MULTIPOLYGON MUTEX MYSQL_ERRNO
NAME NAMES NATIONAL NCHAR NDB NDBCLUSTER NEVER NEW NEXT NO NODEGROUP
    NONBLOCKING NONE NO_WAIT NUMBER NVARCHAR
OFFSET OLD_PASSWORD ONE ONLY OPEN OPTIONS OWNER
PACK_KEYS PAGE PARSER PARSE_GCOL_EXPR PARTIAL PARTITIONING PARTITIONS PASSWORD
    PHASE PLUGIN PLUGINS PLUGIN_DIR POINT POLYGON PORT PRECEDES PREPARE
    PRESERVE PREV PRIVILEGES PROCESSLIST PROFILE PROFILES PROXY
QUARTER QUERY QUICK
READ_ONLY REBUILD RECOVER REDOFILE REDO_BUFFER_SIZE REDUNDANT RELAY RELAYLOG
    RELAY_LOG_FILE RELAY_LOG_POS RELAY_THREAD RELOAD REMOVE REORGANIZE REPAIR
    REPEATABLE REPLICATE_DO_DB REPLICATE_DO_TABLE REPLICATE_IGNORE_DB
    REPLICATE_IGNORE_TABLE REPLICATE_REWRITE_DB REPLICATE_WILD_DO_TABLE
    REPLICATE_WILD_IGNORE_TABLE REPLICATION RESET RESTORE RESUME
    RETURNED_SQLSTATE RETURNS REVERSE ROLLBACK ROLLUP ROTATE ROUTINE ROW ROWS
    ROW_COUNT ROW_FORMAT RTREE
SAVEPOINT SCHEDULE SCHEMA_NAME SECOND SECURITY SERIAL SERIALIZABLE SERVER
    SESSION SHARE SHUTDOWN SIGNED SIMPLE SLAVE SLOW SNAPSHOT SOCKET SOME SONAME
    SOUNDS SOURCE SQL_AFTER_GTIDS SQL_AFTER_MTS_GAPS SQL_BEFORE_GTIDS
    SQL_BUFFER_RESULT SQL_CACHE SQL_NO_CACHE SQL_THREAD SQL_TSI_DAY
    SQL_TSI_HOUR SQL_TSI_MINUTE SQL_TSI_MONTH SQL_TSI_QUARTER SQL_TSI_SECOND
    SQL_TSI_WEEK SQL_TSI_YEAR STACKED START STARTS STATS_AUTO_RECALC
    STATS_PERSISTENT STATS_SAMPLE_PAGES STATUS STOP STORAGE STRING
    SUBCLASS_ORIGIN SUBJECT SUBPARTITION SUBPARTITIONS SUPER SUSPEND SWAPS
    SWITCHES
TABLES TABLESPACE TABLE_CHECKSUM TABLE_NAME TEMPORARY TEMPTABLE TEXT THAN TIME
    TIMESTAMP TIMESTAMPADD TIMESTAMPDIFF TRANSACTION TRIGGERS TRUNCATE TYPE
    TYPES
UNCOMMITTED UNDEFINED UNDOFILE UNDO_BUFFER_SIZE UNICODE UNINSTALL UNKNOWN UNTIL
    UPGRADE USER USER_RESOURCES USE_FRM
VALIDATION VALUE VARIABLES VIEW WAIT
WARNINGS WEEK WEIGHT_STRING WITHOUT WORK WRAPPER
X509 XA XID XML
YEAR
    """
    mysql_keywords = sorted(list(set(
        mysql_reserved_words.split() +
        ANSI92_RESERVED_WORD_LIST.split()
    )))
    # log.critical(mysql_keywords)
    keyword = make_words_regex(mysql_keywords, caseless=True, name="keyword")

    # -------------------------------------------------------------------------
    # Comments
    # -------------------------------------------------------------------------
    # http://dev.mysql.com/doc/refman/5.7/en/comments.html
    comment = (ansi_comment | bash_comment | cStyleComment)

    # -----------------------------------------------------------------------------
    # identifier
    # -----------------------------------------------------------------------------
    # http://dev.mysql.com/doc/refman/5.7/en/identifiers.html
    bare_identifier_word = make_regex_except_words(
        r"\b[a-zA-Z0-9$_]*\b",
        mysql_keywords,
        caseless=True,
        name="bare_identifier_word"
    )
    liberal_identifier_word = make_pyparsing_regex(
        r"\b[a-zA-Z0-9$_]*\b",
        caseless=True,
        name="liberal_identifier_word"
    )
    identifier = (
        bare_identifier_word |
        QuotedString(quoteChar="`", unquoteResults=False)
    ).setName("identifier")
    liberal_identifier = (
        liberal_identifier_word |
        QuotedString(quoteChar="`", unquoteResults=False)
    ).setName("liberal_identifier")
    # http://dev.mysql.com/doc/refman/5.7/en/charset-collate.html
    collation_name = identifier.copy()
    column_name = identifier.copy()
    column_alias = identifier.copy()
    table_name = identifier.copy()
    table_alias = identifier.copy()
    index_name = identifier.copy()
    function_name = identifier.copy()
    parameter_name = identifier.copy()
    database_name = identifier.copy()
    partition_name = identifier.copy()

    no_dot = NotAny('.')
    # MySQL allows keywords in the later parts of combined identifiers;
    # therefore, for example, "count.thing.thing" is not OK, but
    # "thing.thing.count" is.
    table_spec = (
        Combine(database_name + '.' + liberal_identifier + no_dot) |
        table_name + no_dot
    ).setName("table_spec")
    column_spec = (
        Combine(database_name + '.' + liberal_identifier + '.' +
                liberal_identifier + no_dot) |
        Combine(table_name + '.' + liberal_identifier + no_dot) |
        Combine(column_name + no_dot)
    ).setName("column_spec")

    # http://dev.mysql.com/doc/refman/5.7/en/expressions.html
    bind_parameter = Literal('?')

    # http://dev.mysql.com/doc/refman/5.7/en/user-variables.html
    variable = Regex(r"@[a-zA-Z0-9\.$_]+").setName("variable")

    # http://dev.mysql.com/doc/refman/5.7/en/functions.html
    argument_list = (
        delimitedList(expr).setName("arglist").setParseAction(', '.join)
    )
    # ... we don't care about sub-parsing the argument list, so use combine=True
    # or setParseAction: http://stackoverflow.com/questions/37926516
    function_call = Combine(function_name + LPAR) + argument_list + RPAR

    # http://dev.mysql.com/doc/refman/5.7/en/partitioning-selection.html
    partition_list = (
        LPAR + delim_list(partition_name, combine=True) + RPAR
    ).setName("partition_list")

    # http://dev.mysql.com/doc/refman/5.7/en/index-hints.html
    index_list = delim_list(index_name, combine=False)
    # ... see pyparsing_bugtest_delimited_list_combine
    index_hint = (
        (
            USE + (INDEX | KEY) +
            Optional(FOR + (JOIN | (ORDER + BY) | (GROUP + BY))) +
            LPAR + Optional(index_list) + RPAR
        ) |
        (
            IGNORE + (INDEX | KEY) +
            Optional(FOR + (JOIN | (ORDER + BY) | (GROUP + BY))) +
            LPAR + index_list + RPAR
        ) |
        (
            FORCE + (INDEX | KEY) +
            Optional(FOR + (JOIN | (ORDER + BY) | (GROUP + BY))) +
            LPAR + index_list + RPAR
        )
    )
    index_hint_list = delim_list(index_hint, combine=True).setName(
        "index_hint_list")

    # -----------------------------------------------------------------------------
    # CASE
    # -----------------------------------------------------------------------------
    # NOT THIS: https://dev.mysql.com/doc/refman/5.7/en/case.html
    # THIS: https://dev.mysql.com/doc/refman/5.7/en/control-flow-functions.html#operator_case  # noqa
    case_expr = (
        (
            CASE + expr +
            OneOrMore(WHEN + expr + THEN + expr) +
            Optional(ELSE + expr) +
            END
        ) | (
            CASE +
            OneOrMore(WHEN + expr + THEN + expr) +
            Optional(ELSE + expr) +
            END
        )
    ).setName("case_expr")

    # -------------------------------------------------------------------------
    # MATCH
    # -------------------------------------------------------------------------
    # https://dev.mysql.com/doc/refman/5.7/en/fulltext-search.html#function_match
    search_modifier = (
        (IN + NATURAL + LANGUAGE + MODE + Optional(WITH + QUERY + EXPANSION)) |
        (IN + BOOLEAN + MODE) |
        (WITH + QUERY + EXPANSION)
    )
    match_expr = (
        MATCH + LPAR + delim_list(column_spec) + RPAR +
        AGAINST + LPAR + string_literal + Optional(search_modifier) + RPAR
    ).setName("match_expr")
    # ... don't use "expr"; MATCH AGAINST uses restricted expressions, and we
    # don't want it to think that "MATCH ... AGAINST ('+keyword' IN
    # BOOLEAN MODE)" resembles the IN in "WHERE something IN (SELECT ...)"

    # -----------------------------------------------------------------------------
    # Expressions
    # -----------------------------------------------------------------------------
    # http://dev.mysql.com/doc/refman/5.7/en/expressions.html
    # https://pyparsing.wikispaces.com/file/view/select_parser.py
    # http://dev.mysql.com/doc/refman/5.7/en/operator-precedence.html
    expr_term = (
        INTERVAL + expr + time_unit |
        # "{" + identifier + expr + "}" |  # see MySQL notes; antique ODBC syntax  # noqa
        Optional(EXISTS) + LPAR + select_statement + RPAR |
        # ... e.g. mycol = EXISTS(SELECT ...)
        # ... e.g. mycol IN (SELECT ...)
        LPAR + delim_list(expr) + RPAR |
        # ... e.g. mycol IN (1, 2, 3) -- "(1, 2, 3)" being a term here
        case_expr |
        match_expr |
        bind_parameter |
        variable |
        function_call |
        literal_value |
        column_spec  # not just identifier
    )
    UNARY_OP, BINARY_OP, TERNARY_OP = 1, 2, 3
    expr << infixNotation(expr_term, [
        # https://pythonhosted.org/pyparsing/

        # Having lots of operations in the list here SLOWS IT DOWN A LOT.
        # Just combine them into an ordered list.
        (BINARY | COLLATE | oneOf('! - + ~'), UNARY_OP, opAssoc.RIGHT),
        (
            (
                oneOf('^ * / %') | DIV | MOD |
                oneOf('+ - << >> & | = <=> >= > <= < <> !=') |
                (IS + Optional(NOT)) | LIKE | REGEXP | (Optional(NOT) + IN) |
                (SOUNDS + LIKE)
            ),  # RNC; presumably at same level as LIKE
            BINARY_OP,
            opAssoc.LEFT
        ),
        ((BETWEEN, AND), TERNARY_OP, opAssoc.LEFT),
        # CASE handled above (hoping precedence is not too much of a problem)
        (NOT, UNARY_OP, opAssoc.RIGHT),
        (AND | '&&' | XOR | OR | '||' | ':=', BINARY_OP, opAssoc.LEFT),
    ], lpar=LPAR, rpar=RPAR)
    # ignores LIKE [ESCAPE]

    # -------------------------------------------------------------------------
    # SELECT
    # -------------------------------------------------------------------------

    compound_operator = UNION + Optional(ALL | DISTINCT)
    # no INTERSECT or EXCEPT in MySQL?

    ordering_term = (
        expr + Optional(COLLATE + collation_name) + Optional(ASC | DESC)
    )
    # ... COLLATE can appear in lots of places;
    # http://dev.mysql.com/doc/refman/5.7/en/charset-collate.html

    join_constraint = Optional(Group(  # join_condition in MySQL grammar
        (ON + expr) |
        (USING + LPAR + delim_list(column_name) + RPAR)
    ))

    # http://dev.mysql.com/doc/refman/5.7/en/join.html
    join_op = Group(
        COMMA |
        STRAIGHT_JOIN |
        NATURAL + (Optional(LEFT | RIGHT) + Optional(OUTER)) + JOIN |
        (INNER | CROSS) + JOIN |
        Optional(LEFT | RIGHT) + Optional(OUTER) + JOIN
        # ignores antique ODBC "{ OJ ... }" syntax
    )

    join_source = Forward()
    single_source = (
        (
            table_spec.copy().setResultsName("from_tables",
                                             listAllMatches=True) +
            Optional(PARTITION + partition_list) +
            Optional(Optional(AS) + table_alias) +
            Optional(index_hint_list)
        ) |
        (select_statement + Optional(AS) + table_alias) +
        (LPAR + join_source + RPAR)
    )
    join_source << Group(
        single_source + ZeroOrMore(join_op + single_source + join_constraint)
    )("join_source")
    # ... must have a Group to append to it later, it seems
    # ... but name it "join_source" here, or it gets enclosed in a further list
    #     when you name it later

    aggregate_function = (
        # https://dev.mysql.com/doc/refman/5.7/en/group-by-functions.html
        AVG |
        BIT_AND |
        BIT_OR |
        BIT_XOR |
        COUNT |  # also: special handling for COUNT(DISTINCT ...), see below
        GROUP_CONCAT |
        MAX |
        MIN |
        STD |
        STDDEV |
        STDDEV_POP |
        STDDEV_SAMP |
        SUM |
        VAR_POP |
        VAR_SAMP |
        VARIANCE
    )
    result_base = (
        # Aggregate functions: e.g. "MAX(" allowed, "MAX (" not allowed
        Combine(COUNT + LPAR) + '*' + RPAR |  # special aggregate function
        Combine(COUNT + LPAR) + DISTINCT + expr + RPAR |  # special aggregate function  # noqa
        Combine(aggregate_function + LPAR) + expr + RPAR |
        expr |
        '*' |
        Combine(table_name + '.' + '*') |
        column_spec |
        literal_value
    )
    result_column = (
        result_base + Optional(Optional(AS) + column_alias)
    ).setResultsName("select_columns", listAllMatches=True)

    # -------------------------------------------------------------------------
    # SELECT
    # -------------------------------------------------------------------------
    # http://dev.mysql.com/doc/refman/5.7/en/select.html
    """
    SELECT
        [ALL | DISTINCT | DISTINCTROW ]
          [HIGH_PRIORITY]
          [MAX_STATEMENT_TIME = N]
          [STRAIGHT_JOIN]
          [SQL_SMALL_RESULT] [SQL_BIG_RESULT] [SQL_BUFFER_RESULT]
          [SQL_CACHE | SQL_NO_CACHE] [SQL_CALC_FOUND_ROWS]
        select_expr [, select_expr ...]
        [FROM table_references
          [PARTITION partition_list]
        [WHERE where_condition]
        [GROUP BY {col_name | expr | position}
          [ASC | DESC], ... [WITH ROLLUP]]
        [HAVING where_condition]
        [ORDER BY {col_name | expr | position}
          [ASC | DESC], ...]
        [LIMIT {[offset,] row_count | row_count OFFSET offset}]

    ... ignore below here...

        [PROCEDURE procedure_name(argument_list)]
        [INTO OUTFILE 'file_name'
            [CHARACTER SET charset_name]
            export_options
          | INTO DUMPFILE 'file_name'
          | INTO var_name [, var_name]]
        [FOR UPDATE | LOCK IN SHARE MODE]]
    """
    where_expr = Group(expr).setResultsName("where_expr")
    where_clause = Group(
        Optional(WHERE + where_expr)
    ).setResultsName("where_clause")
    select_core = (
        SELECT +
        Group(Optional(ALL | DISTINCT | DISTINCTROW))("select_specifier") +
        Optional(HIGH_PRIORITY) +
        Optional(MAX_STATEMENT_TIME + '=' + integer) +
        Optional(STRAIGHT_JOIN) +
        Optional(SQL_SMALL_RESULT) +
        Optional(SQL_BIG_RESULT) +
        Optional(SQL_BUFFER_RESULT) +
        Optional(SQL_CACHE | SQL_NO_CACHE) +
        Optional(SQL_CALC_FOUND_ROWS) +
        Group(delim_list(result_column))("select_expression") +
        Optional(
            FROM + join_source +
            Optional(PARTITION + partition_list) +
            where_clause +
            Optional(
                GROUP + BY +
                delim_list(ordering_term +
                           Optional(ASC | DESC))("group_by_term") +
                Optional(WITH + ROLLUP)
            ) +
            Optional(HAVING + expr("having_expr"))
        )
    )
    select_statement << (
        select_core +
        ZeroOrMore(compound_operator + select_core) +
        Optional(
            ORDER + BY +
            delim_list(ordering_term +
                       Optional(ASC | DESC))("order_by_terms")
        ) +
        Optional(LIMIT + (
            (Optional(integer("offset") + COMMA) + integer("row_count")) |
            (integer("row_count") + OFFSET + integer("offset"))
        )) +
        # PROCEDURE ignored
        # rest ignored
        Optional(';')
    )
    select_statement.ignore(comment)

    # http://dev.mysql.com/doc/refman/5.7/en/identifiers.html
    # ... approximately (and conservatively):
    MYSQL_INVALID_FIRST_IF_UNQUOTED = re.compile(r"[^a-zA-Z_$]")
    MYSQL_INVALID_IF_UNQUOTED = re.compile(r"[^a-zA-Z0-9_$]")

    def __init__(self):
        super().__init__()

    @classmethod
    def quote_identifier(cls, identifier: str) -> str:
        return "`{}`".format(identifier)

    @classmethod
    def is_quoted(cls, identifier: str) -> bool:
        return identifier.startswith("`") and identifier.endswith("`")

    @classmethod
    def requires_quoting(cls, identifier: str) -> bool:
        assert identifier, "Empty identifier"
        if cls.MYSQL_INVALID_IF_UNQUOTED.search(identifier):
            return True
        firstchar = identifier[0]
        if cls.MYSQL_INVALID_FIRST_IF_UNQUOTED.search(firstchar):
            return True
        if identifier.upper() in cls.mysql_keywords:
            return True
        return False

    @classmethod
    def get_grammar(cls):
        # Grammar (here, just SELECT)
        return cls.select_statement

    @classmethod
    def get_column_spec(cls):
        return cls.column_spec

    @classmethod
    def get_result_column(cls):
        return cls.result_column

    @classmethod
    def get_join_op(cls):
        return cls.join_op

    @classmethod
    def get_table_spec(cls):
        return cls.table_spec

    @classmethod
    def get_join_constraint(cls):
        return cls.join_constraint

    @classmethod
    def get_select_statement(cls):
        return cls.select_statement

    @classmethod
    def get_expr(cls):
        return cls.expr

    @classmethod
    def get_where_clause(cls):
        return cls.where_clause

    @classmethod
    def get_where_expr(cls):
        return cls.where_expr

    @classmethod
    def test_dialect_specific_1(cls):
        log.info("Testing MySQL-specific aspects (1/2)...")
        test_fail(cls.case_expr, "one two three four")
        test_fail(cls.match_expr, "one two three four")
        test_fail(cls.bind_parameter, "one two three four")
        test_fail(cls.variable, "one two three four")
        test_fail(cls.function_call, "one two three four")
        test_fail(literal_value, "one two three four")
        # test_fail(cls.column_spec, "one two three four")  # matches "one"

    @classmethod
    def test_dialect_specific_2(cls):
        log.info("Testing MySQL-specific aspects (2/2)...")

        log.info("Testing expr")
        test_succeed(cls.expr, "a DIV b")
        test_succeed(cls.expr, "a MOD b")

        log.info("Testing quoted identifiers")
        test_succeed(cls.identifier, "`a`")
        test_succeed(cls.identifier, "`FROM`")
        test_succeed(cls.identifier, "`SELECT FROM`")
        # MySQL uses up to: schema.table.column
        test_succeed(cls.table_spec, "mydb.`my silly table`")
        test_succeed(cls.table_spec, "myschema.mytable")
        test_fail(cls.table_spec, "mydb.myschema.mytable")
        # ... but not 4:
        test_succeed(cls.column_spec, "`my silly table`.`my silly column`")
        test_succeed(cls.column_spec, "myschema.mytable.mycol")
        test_succeed(cls.column_spec, "starfeeder.mass_event.thing")
        test_succeed(cls.column_spec, "starfeeder.mass_event.at")
        test_fail(cls.column_spec, "mydb.myschema.mytable.mycol")

        log.info("Testing variable")
        test_succeed(cls.variable, "@myvar")

        log.info("Testing argument_list")
        test_succeed(cls.argument_list, "@myvar, 5")

        log.info("Testing function_call")
        test_succeed(cls.function_call, "myfunc(@myvar, 5)")

        log.info("Testing index_list")
        test_succeed(cls.index_list, "idx1, idx2")

        log.info("Testing index_hint")
        test_succeed(cls.index_hint, "USE INDEX FOR JOIN (idx1, idx2)")

        log.info("Testing case_expr")
        test_succeed(cls.case_expr, """
            CASE v
              WHEN 2 THEN x
              WHEN 3 THEN y
              ELSE -99
            END
        """)

        log.info("Testing match_expr")
        test_succeed(cls.match_expr, """
             MATCH (content_field)
             AGAINST('+keyword1 +keyword2')
        """)
        test_succeed(cls.match_expr, """
             MATCH (content_field)
             AGAINST('+keyword1 +keyword2' IN BOOLEAN MODE)
        """)


# =============================================================================
# main
# =============================================================================

def main() -> None:
    log.info("TESTING MYSQL DIALECT")
    mysql = SqlGrammarMySQL()
    mysql.test()
    log.info("ALL TESTS SUCCESSFUL")


if __name__ == '__main__':
    main_only_quicksetup_rootlogger()
    main()
